//----- bcm.c ----- 04/02/09 -----(xx/xx/xx)-----

//=============================================================================
//
//  2006-06-27 - Mark Davis
//               - ported skeleton of older DC version of code to new RTOS
//
//=============================================================================

#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#include <sys/nsclSerialProto.h>

#include <sys/farprintf.h>


//--- the 8 least significant bits are for appl/cmd-specific events ---
#define  ProcessBits       0x01L
#define  PacketsBits       0x02L

#define  ShowProcess       (DiagFlags & ProcessBits)
#define  ShowPackets       (DiagFlags & PacketsBits)


#define  SERIAL_BUF_SIZE  64

#define NUM_EXTRA 7
#define OLD_NUM_EXTRA  4


//===== structures for serial cmd data and responses =====
// *** WARNING ***
// Remember that the host expects data in MSB 1st format while the Rabbit
// processor uses LSB 1st format for multi-byte formats
// *** WARNING ***

typedef struct {
  unsigned long rom_date;
  unsigned char dev_type;
} VersionInfo;

// Message format structures
typedef struct {
  unsigned char  Reading[4];     // scaled current reading
  unsigned char  Range;          // current range
  unsigned short Status;         // status flags
} PollMsgResp;

typedef struct {
  unsigned char  Range;          // current range to be used
  unsigned char  State;          // turns filter/frontend on or off
} SetMsgData;

typedef struct {
  char           SwitchNum;      // switch number (0, 1 or 2)
  unsigned char  Current[4];     // for 1,2: current being applied externally
} ADCCalMessageForm;


//----- definitions for access to ModBus-type registers via serial cmds ---
#define  NumModRegs  4
typedef struct ModBusMsgStruct  {  // for NSCL serial interface only...
  uchar   modBusCmd;
  ushort  firstRegNum;
  uchar   numRegs;
  ushort  regValue[NumModRegs];
} ModBusMsgData;

//--- values for ->modBusCmd ---
#define  MBReadRegs    1
#define  MBWriteRegs   2
#define  MBReadRORegs  3
#define  MBWriteCoils  4


//--- offsets in to array of error counters ---
#define CharCount    0
#define CheckSum     1
#define DataChar     2
#define Framing      3
#define InvalidSrc   4
#define Overrun      5
#define Parity       6
#define Reset        7
#define EndNotCR     8
#define TimeOut      9
#define UndefCmd    10
//--------------------
#define NumErrTypes 11


ushort errCounts[NumErrTypes];


//=============================================================================
//  Device-specific functions to process serial commands
//=============================================================================

#ifdef NSCLSerial

//-----------------------------------------------------------------------------
//  Debug command processing
//-----------------------------------------------------------------------------
char Cmd_Debug_Proc(char DevNum, char *data, int *datalen)
{
  long d;
  int ii;

  for (ii = 0; ii < OLD_NUM_EXTRA; ii++) {
    d = DecimalRep(BCM[DevNum].Extra[ii]);
    memmove(data+(ii*4),&d,4);
  }
  (*datalen) = OLD_NUM_EXTRA * 4;
  return(STAT_OK);
}

//-----------------------------------------------------------------------------
//  Calibrate command processing
//-----------------------------------------------------------------------------
char Cmd_Calibrate_Proc(char DevNum)
{
  BCM_StartCalibrate(DevNum,true);  // request a full calibrate
  return(STAT_OK);
}

//-----------------------------------------------------------------------------
//  Init command processing
//-----------------------------------------------------------------------------
char Cmd_Init_Proc(char devNum)
{
  InitCmds(devNum);  return(STAT_OK);
}

//-----------------------------------------------------------------------------
//  Poll command processing
//-----------------------------------------------------------------------------
char Cmd_Poll_Proc(char DevNum, PollMsgResp *pollData)
{
  float fTemp;

  // fill in current field
  fTemp = BCM[DevNum].avgCur;
  reverseString((char *)&fTemp, sizeof(float));
  memcpy(pollData->Reading, &fTemp, sizeof(float));

  // fill in range field
  pollData->Range = (unsigned char) BCM[DevNum].Range.ActiveSetting;

  // fill in status field
  pollData->Status = BCM_GetStatus(DevNum);
  reverseString((char *)&pollData->Status, sizeof(unsigned short));

  return(STAT_OK);
}

//-----------------------------------------------------------------------------
//  Set command processing
//-----------------------------------------------------------------------------
char Cmd_Set_Proc(char DevNum, SetMsgData* setData)
{
  unsigned char newRange;
  unsigned short aidx;

  // extract range
  newRange = setData->Range;

  // force range to be 0..NumRanges
  if (newRange < 0) {
    newRange = 0;
  } else if (newRange >= NumRanges) {
    newRange = NumRanges - 1;
  }

  // get AVG_Bx bits out of state byte
  aidx = 0;
  aidx |= (setData->State & SETSTATE_AVG_B0) ? 0x01 : 0x00;
  aidx |= (setData->State & SETSTATE_AVG_B1) ? 0x02 : 0x00;
  aidx |= (setData->State & SETSTATE_AVG_B2) ? 0x04 : 0x00;

  // request user settings
  BCM_UserSetFrontEnd(DevNum, BOOL(setData->State & SETSTATE_FRONTEND) );
  BCM_UserSetFilter(DevNum, BOOL(setData->State & SETSTATE_FILTER) );
#ifdef BCM_SS
  BCM_UserSetTestCur(DevNum, BOOL(setData->State & SETSTATE_TESTCUR) );
#endif
  BCM_UserSetRange(DevNum, newRange);
  BCM_UserSetAvgSizeIdx(DevNum, aidx);

  return (STAT_OK);
}

//-----------------------------------------------------------------------------
//  Extended-calibrate routines, selected by SwitchNum field of adcData
//
//  SwitchNum 0 : zero offsets (until next calibrate, power-cycle or reboot)
//            1 : point 1 for ADC calibration
//            2 : point 2 for ADC calibration
//
//  NOTE:  This routine is out of date as the BCM2002 calibrates each
//         range independently
//-----------------------------------------------------------------------------
char Cmd_XCal_Proc(char DevNum, ADCCalMessageForm *adcData)
{
  float fTemp;

  // extract floating-point number
  memcpy(&fTemp,adcData->Current,sizeof(float));
  reverseString((char *)&fTemp,sizeof(float));

  // extract switch number, 0, 1 or 2
  switch (adcData->SwitchNum) {
    case '0':
        // zero BCM calibration
        BCM_ClearOffsets(DevNum);
        break;
    case '1':
        // point 1 for ADC calibration
//        BCM_ADCCalibrationPoint(DevNum,1,fTemp);
        break;
    case '2':
        // point 2 for ADC calibration
//        BCM_ADCCalibrationPoint(DevNum,2,fTemp);
        break;
  }

  return(STAT_OK);
}

//-----------------------------------------------------------------------------
//  Set-and-Poll command processing
//-----------------------------------------------------------------------------
char Cmd_SetAndPoll_Proc
     (char           DevNum,
      PollMsgResp   *pollData,
      SetMsgData    *setData)
{
  char status;

  status = Cmd_Set_Proc(DevNum, setData);
  if (status == STAT_OK) status = Cmd_Poll_Proc(DevNum, pollData);

  return(status);
}

#endif


//-----------------------------------------------------------------------------
//  Take the message in the RS485 input buffer and verify its correctness.
//  If the message is not for this device, ignore it.  If it is for this dev,
//  and it is both correct and intact, perform the command.  Otherwise inform
//  the master of the error.
//-----------------------------------------------------------------------------
int ProcessNSCLMsg(char *cmdBuf, char *replyBuffer)
{
  char  status;
  uchar  command, ID, dataCount, DevNum;
  int  i, respLen;  //, datalen;
  VersionInfo  *vi;
  char  cmdData[SERIAL_BUF_SIZE], respData[SERIAL_BUF_SIZE];


  *replyBuffer = 0;  ID = cmdBuf[1] - '@';

//--- if msg is not for one of our 2 addresses, skip further processing ---
  if (ID == (deviceID & 0x0f))
    DevNum = 0;
  else if (ID == ((deviceID +1) & 0x0f))
    DevNum = 1;
  else
    return 1;

  if (ShowPackets)
    printf("for ID:0x%02X (dev#%d), [%.*s]\r\n",
           ID, DevNum, strlen(cmdBuf)-2, cmdBuf+1);

//--- initialize the variables ---
  ID = (char) 0x7F;

//--- see if a valid message is waiting, skip further processing if not ---
  status = checkMessage(cmdBuf, &ID);
//printf("checkMessage() returned: 0x%02X(%c)\r\n", status, status);  //tdebug
  if (status != STAT_OK) {
    switch (status)  {
      case STAT_BADFORMAT:         ++errCounts[DataChar];   break;
      case STAT_CHECKSUM_ERROR:    ++errCounts[CheckSum];   break;
      case STAT_BAD_FIELD_CHAR:    ++errCounts[DataChar];   break;
      case STAT_MESSAGE_LONG:      ++errCounts[CharCount];  break;
      case STAT_MESSAGE_SHORT:     ++errCounts[CharCount];  break;
      case STAT_UNDEFINED_COMMAND: ++errCounts[UndefCmd];   break;
      case STAT_NOT_FROM_MASTER:   return 2;  // ignore echo of our own response
    }
    if (ShowErrors)
      printf("serial pkt error: status:%c for ID:%d\r\n", status, ID);
    return -1;
  }

//--- if everything still ok, determine the message type and process it ---
  unpackMessage(cmdBuf, &command, cmdData, &dataCount);

  status = STAT_OK;  //tdebug

  switch (command) {

    case CMD_INIT:
      if (ShowPackets)  printf("(Init)\r\n");
//      status = Cmd_Init_Proc(DevNum);
      vi = (VersionInfo *) respData;
      vi->dev_type = devType;
      vi->rom_date = 0x20060623L;  //finish
      reverseString(respData, 5);  respLen = 5 * 2;
      break;


    case CMD_CLEAR_ERR:
      if (ShowPackets)  printf("(ClearErrs)\r\n");  //tdebug
      for (i=0; i<NumErrTypes; ++i)  errCounts[i] = 0;
      status = STAT_OK;  respLen = 0;
      break;


    case CMD_DEBUG:
      if (ShowPackets)  printf("(Debug)\r\n");  //tdebug
//      status = Cmd_Debug_Proc(DevNum, respData, &datalen);
//      respLen = datalen * 2;
      respLen = 16;  //tdebug
      break;


    case CMD_CALIBRATE:
      if (ShowPackets)  printf("(Calib)\r\n");  //tdebug
//      status = Cmd_Calibrate_Proc(DevNum);
      respLen = 0;
      break;


    case CMD_MODBUSREGS:
       if (ShowPackets)  printf("(ModRegs)\r\n");  //tdebug
//      status = Cmd_ModBusRegs_Proc((ModBusMsgData *) cmdData,
//                                   (ModBusMsgData *) respData, lp );
      respLen = sizeof(ModBusMsgData) * 2;
      break;


    case CMD_POLL:
      if (ShowPackets)  printf("(Poll)\r\n");  //tdebug
//      status = Cmd_Poll_Proc(DevNum,(PollMsgResp *)respData);
      respLen = sizeof(PollMsgResp) * 2;
      break;


    case CMD_SET:
      if (ShowPackets)  printf("(Set)\r\n");  //tdebug
//      status = Cmd_Set_Proc(DevNum,(SetMsgData *)cmdData);
      respLen = 0;
      break;


    case CMD_SET_AND_POLL:
      if (ShowPackets)  printf("(Set+Poll)\r\n");  //tdebug
//      status = Cmd_SetAndPoll_Proc(DevNum,
//                                   (PollMsgResp *) respData,
//                                   (SetMsgData *) cmdData);
      respLen = sizeof(PollMsgResp) * 2;
      break;


    case CMD_XCAL:
      if (ShowPackets)  printf("(XCalib)\r\n");  //tdebug
//      status = Cmd_XCal_Proc(DevNum,(ADCCalMessageForm *)cmdData);
      respLen = 0;
      break;


    case CMD_REBOOT:
      if (ShowPackets)  printf("(Reboot)\r\n");  //tdebug
//      Reboot(30);
      break;


    default:
      if (ShowErrors)  printf("unknown command\r\n");  //tdebug
      status = STAT_UNDEFINED_COMMAND;

  }  // end switch

  if (status == STAT_OK)
    packMessage(replyBuffer, ID, status, respData, respLen);
  else
    packMessage(replyBuffer, ID, status, "", 0);

  return 0;
}


//-----------------------------------------------------------------------------
//  server to process NSCL serial commands
//-----------------------------------------------------------------------------
void _task nsclSeriald(int argc, char **argv)
{
  fd_t  devfd;
  char  pktStartChar;
  uint  i, asyncFmt;
  int  stat;
  ulong  baudRate;
  char  cmdBuf[64], replyBuf[128];


  if (argc < 3)  { printf("Usage: nsclSerald device speed [fmt]\r\n\n");  Exit(-1); }

  if ((devfd = open(argv[1], O_RDWR, 0)) < 0)  {
    printf("Unable to open device: %s\r\n", argv[1]);  Exit(-2);  }

  baudRate = atol(argv[2]);  ioctl(devfd, IOC_SET_BAUD, &baudRate);

//--- check for byte format other than default of 8E1 ---
  if (argc > 3)  for (;;)  {
    asyncFmt = 0;
    if (stricmp(argv[3], "8N1") == 0)  break;
    if (stricmp(argv[3], "7O1") == 0)  { asyncFmt = ASYNC_FMT_7O1;  break; }
    if (stricmp(argv[3], "7E1") == 0)  { asyncFmt = ASYNC_FMT_7E1;  break; }
    if (stricmp(argv[3], "8O1") == 0)  { asyncFmt = ASYNC_FMT_8O1;  break; }
    if (stricmp(argv[3], "8E1") == 0)  { asyncFmt = ASYNC_FMT_8E1;  break; }
    printf("Invalid async format: %s\r\n", argv[3]);
    break;
  }
  if (asyncFmt)  ioctl(devfd, IOC_SET_ASYNC_FMT, &asyncFmt);


//--- set serial driver to use packet mode, waiting for a LF char to ---
//--- indicate the start of each packet ---
  pktStartChar = 0x0a;  // NSCL packets start with a LF
  ioctl(devfd, IOC_SET_PKTMODE, &pktStartChar);

  for (;;)  {
  //--- wait for next packet ---
    for (i=1; i<sizeof(cmdBuf)-1; ++i)  {
      if ((stat = read(devfd, cmdBuf+i, 1)) < 1)  break;
      cmdBuf[i] &= 0x7f;
      if (cmdBuf[i] == 0x0D)  { ioctl(devfd, IOC_END_PKT_RX, NULL);  break; }
    }
    if (stat < 1)  { perror("error reading serial packet");  break; }
    if (cmdBuf[i] != 0x0D)  { perror("incomplete packet");  continue; }

  //--- put the LF back at the start, and add a string terminator ---
    cmdBuf[0] = 0x0A;  cmdBuf[i+1] = 0;

  //--- process the request ---
    stat = ProcessNSCLMsg(cmdBuf, replyBuf);

  //--- send the response ---
    if (stat == 0)  {
      if (ShowPackets)
        printf("response: [%.*s]\r\n", strlen(replyBuf)-2, replyBuf+1);
      ioctl(devfd, IOC_START_PKT_TX, NULL);
      write(devfd, replyBuf, strlen(replyBuf));
      ioctl(devfd, IOC_END_PKT_TX, NULL);
    }
  }

  Exit(0);
}

//-----------------------------------------------------------------------------


