//----- LLRF2_Appl.h ----- 11/14/12 -----(03/05/06)----

//-----------------------------------------------------------------------------
//  Declarations and definitions for things in the application module that
//  the hardware-specific counterpart needs to know about.
//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2012-11-14 - Mark Davis
//               - Added typical and percNoise values to RD_DSP_REG struct to
//                 provide more reasonable values when running on other hdw
//                 platforms.
//
//  2012-01-16 - Mark Davis
//               - Changed argument for releaseLock() to uint
//
//  2011-09-16 - Mark Davis
//               - Created by extracting pieces from the LLRF2_Appl.c file.
//
//-----------------------------------------------------------------------------


#ifndef _LLRF2_Appl_H_
#define _LLRF2_Appl_H_


#include <common.h>

#include <unistd.h>
#include <sys/types.h>

#include <linux/wait.h>

#include <semaphore.h>


//=============================================================================
//  Data structures for managing communication with the DSP and for keeping
//  ModBus register values in sync with the corresponding DSP registers.
//=============================================================================

typedef struct {
  ulong    id;         // DSP "address"
  uchar    type;       // int32, float, bits, ...
  RegInfo *mbReg;      // corresponding Modbus register
  float   *gain;       // ptr to gain value or NULL
  float   *offset;     // ptr to offset value or NULL
  ulong    typical;    // for testing on other hdw
  uint     percNoise;  // max noise as percentage of typical value
} RD_DSP_REG;


typedef struct {
  ulong    id;         // reg # in DSP to write value to (if applicable)
  uchar    type;       // int32, float, bits, ...
  RegInfo *mbReg;      // Modbus register info
  float    minVal;     // minimum value allowed
  float    maxVal;     // maximum value allowed
  uchar    writeMode;  // how to handle writes
  float   *gain;       // ptr to gain value or NULL
  float   *offset;     // ptr to offset value or NULL
} WR_DSP_REG;




//--- values for WR_DSP_REG.writeMode ---
#define  WM_ONCE    0  // write new value only once at time it was received
#define  WM_EVERY   1  // write cur value every pass

//--- values for RD_DSP_REG.type and WR_DSP_REG.type ---
#define  RT_BITS   DT_U32
#define  RT_S32    DT_S32
#define  RT_U32    DT_U32
#define  RT_F32    DT_F32
#define  RT_PHASE  99


typedef struct  {
  int  count;
  PID  owner;
//  WaitQueue  tasks;
  sem_t  sem;
}  LockData;

enum LockIDS { APPL_LOCK,
               WAVE_DATA_0, WAVE_DATA_1, WAVE_DATA_2, WAVE_DATA_3,
               WAVE_DATA_4, WAVE_DATA_5, WAVE_DATA_6, WAVE_DATA_7,
               WAVE_DATA_8, WAVE_DATA_9,
               NUM_LOCKS };

extern LockData  locks[NUM_LOCKS];

int  getLock(uint id, ulong msTimeout);
int  releaseLock(uint id);

#define LockDepth(id)  ( locks[id].count )
#define LockOwnedBy(id)  ( locks[id].owner )
#define NotLockOwner(id)  (LockOwnedBy(id) != getpid())



extern bool  fpgaInitialized, showFiller;



void prepForReset(void);


char *fmtDSPval(ulong dspVal, ulong regNum, char *buf);


void dsp2mb(RegInfo *mbReg, S16 *mbVal,
            const RD_DSP_REG far *dspReg, ulong *dspVal);

void mb2dsp(RegInfo *mbReg, S16 *mbVal,
            const WR_DSP_REG far *dspReg, ulong *dspVal);


extern const uint Num_RD_DSP_Regs, Num_WR_DSP_Regs;


#endif

