;  Version: 1.70
;
;  C Startup.asm - Control Cross-C all-purpose startup code.
;
;  Copyright (c) 2000-2005 Softools, Inc.  All rights reserved.
;
;  Since this file is self-configuring using several assemble-time
;  prompts, it should be assembled separately for each application
;  program being developed.  Labels can be defined to bypass the
;  prompts - the WinIDE New Project Wizard in fact does this.
;
;  If a pre-startup module is used to initialize other hardware,
;  interrupt vectors, etc., the global entry point CSTART should
;  be jumped to from that module.  Otherwise, this module is placed
;  at location 0 and will handle the startup.  The state of interrupts
;  is NOT changed by this module.  Therefore, if a pre-startup module
;  turns them off, main () or this module should turn them back on.
;  The SP and a MMU and required Rabbit peripherals are initialized
;  by this module.
;
;  The following segments are used in the Rabbit Control Cross-C
;  environment.  Note that not all segments are necessarily used in
;  all applications.
;
; **************************************
; THis has been modified by SHDesigns to run from RAM.
; Two additional defines are needed:
;     RAM_COMPILE - Says run from RAM (loaded by boot manager)
;     FAST_RAM    - says this is a board with FAST RAM. (3200/3300/3800)
;
; **************************************
;
;  Segment      Contains
;  ==================================================================
;  CODE         Startup code, libraries and near compiled code.
;  STARTUP      Startup vector table used to dynamically run startup
;                 initialization code (must be conditionally included).
;  CONST        String literals, const objects and
;                 initializers for automatic aggregates.
;  FARCONST     Constants stored far.
;  DATA         Initialized data area.  Will be copied from ROM to RAM.
;  FARCODE      Code compiled as far functions in C.
;  FARDATA      Initialized far data area.  This is stored out of the
;                 logical address space and not initialized by cstart!
;  BSS          Uninitialized data area.  This will be zeroed on startup.
;  BSSZ         Initialized by 0 data.  This will be zeroed on startup.
;  STACK        The amount reserved for the CPU stack
;               The remmaining area is left over and used for the dynamic heap.
;  FARBSS       Uninitialized far data area.  The startup code DOES NOT
;                 zero this space!
;
;  --------------------------------------------------------------------------
;
;  The following assembly-time questions are asked to configure the
;    application:
;
;  We prompt for the CPU stack size.  This eliminates editing
;    this file just to change the stack size.
;
;  We prompt to find out if the system is battery backed.
;    If it is, then we do not zero-fill BSS and
;    We prompt if the system needs a power-up test for initialization.
;      If it does, then a variable stored in segment DATA is compared
;      to a known default value.  If the variable contains this value,
;      the initialization is skipped.  Otherwise, it is performed.
;    If not, segment BSS is never zero filled.
;
;  If STACK_SIZE is not supplied it will be prompted
;
;-----------------------------------------------------------------------------
Y       equ     1
N       equ     0
;
        .ifdef  STACK_SIZE
;
size    .equ    STACK_SIZE      ; Application stack size?

        .ifdef  BATT_BACKED
        .error  "Battery backing RAM globally is no longer supported - use a new segment"
bb_sys  .equ    N               ; Battery backed system?
        .endif

        .ifndef INC_DEBUG
INC_DEBUG .equ  Y
        .endif
        .if     ! INC_DEBUG
        .printx Building NO DEBUG region C startup (cannot debug in WinIDE)
        .endif

        .ifndef _FLASH_DEBUG
_FLASH_DEBUG    .equ    N
        .endif

        .ifndef GEN_CRC
GEN_CRC .equ    N
        .endif

        .else
;
size::  .input  "Application CPU stack size? "
INC_DEBUG:.input "Include 2k debug kernel region (Y/N)? "
GEN_CRC:.input  "Add CRC table to program (Y/N)? "
        .endif
;
        .include Rabbit.inc
;
__stack_size::  .equ    even size
;
        .cseg                           ; Define implicit segment ordering
        .defseg CONST
        .defseg STARTUP

        .dseg
        .defseg FARCODE
        .defseg FARCONST

        .bseg
        .defseg BSSZ
        .defseg IDATA,data
        .defseg STACK,stack             ; Define STACK segment
        .defseg FARBSS,data
        .defseg XMEMSTART,data   ;;NSCL

        .seg    STACK
_stackMarker::
        .ds     4                       ; A 4-byte marker that can check for stack overruns
        .ds     __stack_size - 4
__stack::.equ   $                       ; Initial SP
;
;;=============================================================================
        .cseg
        .phase  0
        jp      _skip                   ; This is at 0
;
inFlash:.ds     1                       ; Flash marker byte 3 (leave uninitialized)
;
        .org    08h
_reserved:.ds   2                       ; Do not use
_19200TC:.ds    1                       ; Bootstrap leaves 19200 TC here
_GCDR:  .ds     1                       ; And GCDR here.
_TAT4R: .ds     1                       ; And TAT4R here.
        .org    0Eh
__CRC_location::
        .ds     2                       ; Store CRC here at E - initially must be 0
;
;  Debug output and input vectors.  The __stdio FILE pointer
;  is set to these entry points.  In production code, printf
;  calls these per character doing nothing.  In debug mode,
;  the debug kernel redirects these to output to the WinIDE.
;
        .org    10h                     ; Debug in code
        ret
;
        .org    18h                     ; Debug out code
        ret
;
        .org    38h                     ; Debug test input code
__kbhit:xor     a,a
        ret
;
        .if     _FLASH_DEBUG            ; For debugging in flash (do not define!)
;
;  Install default vectors for debugging in flash.  These are
;  not loaded when running in RAM since the debug kernel is in RAM.
;
        .org    50h
        lcall   SWBP##
        .dl     0x4612B9EE              ; Mark this program can flash debug

;
;  Include vector for debug kernel serial handler.
;
        .org    0C0h
        lcall   _dSerIntHandler##
        ret
;
        .endif
;
;  If INC_DEBUG is defined, cstart is built without a 2k hole normally
;  required for the debug kernel.  If you disable this, you cannot run
;  this in the debugger (even if the debug kernel is loaded).
;
;  For debugging in flash, this is still required if you want to switch back to RAM.
;
        .if     INC_DEBUG
;
        .org    800h                    ; 800h is the base above the debug kernel
;
        .else
        .org    0x60+40
        .endif
        .dephase
;
;;=============================================================================
_skip:  ljp     cstart                  ; Continue on with long jump
;
;  idBlock definition, defined as an address
;
_idBlock::      .equ    60h
__idBlockMac1:: .equ    60h+0Bh
__idBlockMac2:: .equ    60h+0Bh+6
;
;  This table defines the area of a program to CRC.  This is done
;  by SLINK automatically upon defining the table and location
;  to store the final CRC.  The code to CRC the program based on
;  this table is included in cstart and run if the label GEN_CRC
;  is non-0.  This makes no assumptions about the ordering and size
;  of segments.  The CRC is stored by SLINK (in all data formats
;  written) at __CRC_location.
;
        .if     GEN_CRC
__CRC_table::                           ; Must have this name to signal SLINK
        .dl     SGAT CODE+810h          ; CRC segment CODE (minus first 16 bytes)
        .dl     SGSZ CODE-810h
        .dl     SGAT CONST              ; CRC CONST
        .dl     SGSZ CONST
        .dl     SGAT DATA               ; CRC DATA
        .dl     SGSZ DATA
        .dl     SGAT FARCODE            ; CRC FARCODE
        .dl     SGSZ FARCODE
        .dl     SGAT FARCONST           ; CRC FARCONST
        .dl     SGSZ FARCONST
        .dl     0                       ; Marks end of table
        .endif                          ; GEN_CRC
;
;  kbhit routine for program to test for input from WinIDE.
;
_kbhit::
        jp      __kbhit
;
        .seg    FARCODE
        .align  8
;;=============================================================================
$cstart::
cstart::
        .debug  { cstart -1 215 cstart end
        ipset   3                       ; Being safe
        ljp     ..next                  ; Set XPC in RAM debug
..next: xor     a,a
        ioi ld  (DATASEG),a     ; must be 0
        ld      iir,a
        ld      sp,__stack
;
;  Initialize the MMU SEGSIZE and STACKSEG to map logical data areas to their
;  physical addresses.
;
;;NSCL        ld      a,(sgst BSS >> 8) & 0F0h
        ld      a,(sgst STACK >> 8) & 0F0h  ;;NSCL
        ioi ld  (SEGSIZE),a
;;NSCL        ld      a,(sgat BSS >> 12) - (sgst BSS >> 12)
        ld      a,(sgat STACK >> 12) - (sgst STACK >> 12)  ;;NSCL
        ioi ld  (STACKSEG),a
;
;
;  Initialize Rabbit peripherals here.  This uses a table to
;  quickly load address and data to initialize the Rabbit processor.
;  Simply extend the table as needed.  The Watchdog is set to 2
;  seconds - plenty of time to get to main from here.
;
        ld      bc,..iTblSize           ; Size
        ld      hl,..iTbl               ; Table
        ld      a,(inFlash)             ; Test for flash, skip MBxCR init if not
        or      a                       ; Is non-0 in flash
        jr      nz,..lp
        ld      bc,..rTblSize           ; Size
        ld      hl,..rTbl               ; Table
..lp:   ld      e,(hl)                  ; Addr
        inc     hl
        ioi ldi
        jp      v,..lp
        or      a                       ; Is non-0 in flash
        jr      z,..inRam
        jr      ..cstart
;
;  Flash only!  MBxCR and chip select initialization.
;
;  This defaults to having the Flash at 00000 and RAM at 80000.
;  Note: These do not update the shadow RAM locations.  If you need
;  to readback any registers used here, either initialize the RAM
;  (e.g. _srMMIDR is the MMIDR shadow location) or use the
;  shadow I/O functions in C to reinitialize them.
;  Note: MBxCR register shadows are initialized.
;
;  Defines for MBxCR registers.  Used here and to initialize the shadows.
;  MB0CR must be 0 wait states or the baud rate constants won't be
;  correct. In flash, this starts up in 0 wait states and later if the speed
;  would be too high doubling, it is changed to 1 wait state.  If 2
;  wait states are needed see set1ws below.
;
;  A 512k flash maps 2 quadrants to one CS (CS0).  This poses a problem
;  for a 2*256k system.  In this case, a runtime check is made and if
;  the ID block shows a non-512k device 0, quadrant 2 (40000) is mapped
;  to CS2 using MB1CRXval.  This setup matches all Z-World 2-flash boards.
;
;  Note: Be careful using wait states on early Rabbit 2000 processors.
;
;  You can define these in the project and override the default EQUs here.
;  You then must supply all 5 definitions or errors will result.
;
WS0             .equ    0C0h    ; MBxCR: 0 ws (+1 wr)
WS1             .equ    080h    ; MBxCR: 1 ws (+1 wr)
WS2             .equ    040h    ; MBxCR: 2 ws (+1 wr)
WS3             .equ    000h    ; MBxCR: 4 ws (+1 wr)
;
        .ifndef MB0CRval
;
          .ifdef  RAM_COMPILE
            .ifdef  FAST_RAM
MB0CRval        .equ    WS0 | 6 ; MB0CR: 0 ws, OE0, WE1, CS2 active,
MB1CRval        .equ    WS0 | 6 ; MB1CR: 0 ws, OE0, WE1, CS2 active,
            .else
MB0CRval        .equ    WS0 | 5 ; MB0CR: 0 ws, OE0, WE1, CS1 active,
MB1CRval        .equ    WS0 | 5 ; MB1CR: 0 ws, OE0, WE1, CS1 active,
            .endif
MB1CRXval       .equ    WS0 | 10; MB1CR: 0 ws, OE0, CS2 active, WP enabled
;
MB2CRval        .equ    WS0 | 5 ; MB2CR: 0 ws, OE1, WE1, CS1 active
MB3CRval        .equ    WS0 | 5 ; MB3CR: 0 ws, OE1, WE1, CS1 active
          .else
MB0CRval        .equ    WS0 | 8 ; MB0CR: 0 ws, OE0, CS0 active, WP enabled
MB1CRval        .equ    WS0 | 8 ; MB1CR: 0 ws, OE0, CS0 active, WP enabled
MB1CRXval       .equ    WS0 | 10; MB1CR: 0 ws, OE0, CS2 active, WP enabled
;
MB2CRval        .equ    WS0 | 5 ; MB2CR: 0 ws, OE1, WE1, CS1 active
MB3CRval        .equ    WS0 | 5 ; MB3CR: 0 ws, OE1, WE1, CS1 active
          .endif
;
        .endif
;
;  MBxCR values for running in Flash.
;
MBXCRvals       .equ    $
                .db     MB0CRval, MB1CRval, MB2CRval, MB3CRval
;
..iTbl:
;
;  Setting STACKSEG: This does one of two things for running in flash.
;    1)  If BSS is linked < 80000h for running in RAM, BSS is relocated to
;       80000h using STACKSEG.
;    2)  If BSS is >= 80000h (linked for full 1MB debugging or running), BSS
;       is left as linked.
;  If any other effect is desired, change the value as needed keeping in mind
;       this is for Flash only.  If this is changed, take note that the start
;       of xmem when running in Flash when FARBSS isn't used uses this offset!
;
;  Be sure to update shadows at ..cstart: below for write-only registers that
;  are not being set to 0!
;
;;NSCL    .ifndef     RAM_COMPILE
;;NSCL        .db     STACKSEG,(((sgat BSS >> 12) >= 0x80) * ((sgat BSS >> 12) - (sgst BSS >> 12))) | (((sgat BSS >> 12) < 0x80) * (80h - (sgst BSS >> 12)))
;;        .db     STACKSEG,(((sgat STACK >> 12) >= 0x80) * ((sgat STACK >> 12) - (sgst STACK >> 12))) | (((sgat STACK >> 12) < 0x80) * (80h - (sgst STACK >> 12)))  ;;NSCL
;;NSCL        .else
;;NSCL        .db     STACKSEG,((sgat BSS >> 12) - (sgst BSS >> 12))
        .db     STACKSEG,((sgat STACK >> 12) - (sgst STACK >> 12))
;;NSCL        .endif
        .db     MB0CR,  MB0CRval
        .db     MB1CR,  MB1CRval
        .db     MB2CR,  MB2CRval
        .db     MB3CR,  MB3CRval
        .db     GCDR,   0       ; Start not doubled
..rTbl:                         ; Here down done in RAM
        .db     GOCR,   0A2h    ; Clk low, status low, /BUFEN low
        .db     GCSR,   8       ; GCSR: no periodic int, main osc no div
        .db     MMIDR,  0       ; MMIDR: normal operation
        .db     MTCR,   00Ch    ; MTCR: 1/2 clock early on OE0 & OE1
        .db     GCM0R,  0       ; Normal spread spectum
        .db     GCM1R,  080h    ; Enabled
        .db     TACSR,  001h    ; TACSR: Enable timer A
        .db     SPCR,   080h    ; SPCR: Disable SMODE pins
        .db     BDCR,   080h    ; Disable RST 28
        .db     WDTCR,  5Ah     ; WDTCR: 2 Seconds to get to main
        .db     WDTTR,  0       ; WDTTR: Enable
;
..iTblSize=($-..iTbl)/2
..rTblSize=($-..rTbl)/2
;
..inRam:
        ld      a,(14h)                 ; Bootstrap MB0CR
        cp      a,0C6h                  ; Running in CS2, 0WS?
        ld      hl,8585h                ; MB2CR & MB3CR 1WS on CS1
        jr      z,..MB2CR
        cp      a,086h                  ; Running in CS2, 1WS?
        jr      nz,..cstart
        ld      hl,4545h                ; MB2CR & MB3CR 2WS on CS1
..Mb2CR:ioi ld  (MB2CR),hl              ; Map in CS1 RAM to 80000
        ld      (16h),hl
..cstart:
        call    initIO                  ; Zero uninitialized Rabbit ports
;
        ld      a,(inFlash)             ; Test for flash, skip init if not
        or      a                       ; Is non-0 in flash
        jr      nz,..flash
        call    initCram                ; Initialize BSS and initialized data
        call    initInRam               ; Initialize when running in RAM
        jr      ..ram
..flash:call    initInFlash             ; Initialize when running in Flash
..ram:
;
;  Read GCSR and store reset code.  Set shadows as initialized in ..iTbl.
;
        ioi ld  a,(GCSR)
        rlca
        rlca
        and     a,3
        ld      (_resetCode),a
        ld      a,1
        ld      (_srTACSR##),a          ; Set TACSR (timer A is running)
        ld      (_srTAPR##),a           ; Default values on a reset
        ld      a,15h
        ld      (_srPCDR##),a
        ld      a,0C0h
        ld      (_srPBDDR##),a
        add     a,a                     ; 80h
        ld      (_srSPCR##),a
        ld      (_srGCM1R##),a          ; As set above in ..iTbl
        ld      a,0A2h
        ld      (_srGOCR##),a
        ld      a,0Ch
        ld      (_srMTCR##),a
;
;  Check for a debug cable connection and enter debug kernel if detected.
;
        .if     _FLASH_DEBUG            ; For debugging in flash (do not define!)
        ioi ld  a,(SPCR)                ; Check for debug cable
        and     40h                     ; SMODE0
        jr      z,..flDbg
        ljp     _dkStart##
..flDbg:
        .endif
;
        lcall   $initXmem               ; Size memory and init Xmem
;
;  Initialize __stdio to the output vectors.
;  Set the __stack_marker to known values to detect stack overruns.
;  Call main ().  Note, main must be far.  If it is declared near,
;  change the lcall to call and $main to _main.
;
        ld      hl,__debugStdio
        ld      (__stdio##),hl
        ld      ix,_stackMarker
        ld      hl,5555h
        ld      (ix),hl
        add     hl,hl                   ; AAAA
        ld      (ix+2),hl
        ld      hl,_stackMarker + 4
        ld      de,_stackMarker + 5
        ld      bc,__stack_size - 5
        ld      (hl),55h
        ldir                            ; Fill stack with 55s
;
        .ifdef  STARTUP_SEG
        call    statupFunctions         ; If enabled, call premain startup function
        .endif
;
        lcall   $main##                 ; Just halt if it returns
        ipset   3
        ld      a,(inFlash)
        or      a,a                     ; If running in RAM do RST
        jr      nz,$                    ; If running in Flash - hang
_mainExit::
        rst     28h                     ; break into debugger
;
;  Here, the following routines are stubbed, since there is little that
;  can be done in an embedded system.  If you DO want to handle these
;  somehow, remove the public entry points here.
;
_abort::
        rst     28h                     ; break into debugger
;
;  Zero Rabbit registers in zTbl
;
initIO: xor     a,a
        ld      b,..zTblSize
        ld      hl,..zTbl
..zlp:  ld      e,(hl)
        ioi ld  (de),a
        inc     hl
        djnz    ..zlp
        ret
;
;  This table of IO addresses will be zeroed.  This is to ensure
;  that uninitialized Rabbit peripherals and registers come up
;  consistently in Flash and RAM execution.
;
..zTbl: .db     PADR, PBDR, PCDR, PDDDR, PDDR, PDCR, PDFR, PDDCR
        .db     PEDDR, PEDR, PECR, PEFR, TBCSR, TBCR, IB0CR, IB1CR
        .db     IB2CR, IB3CR, IB4CR, IB5CR, IB6CR, IB7CR, I0CR, I1CR
        .db     TAT1R, SBCR, SCCR, SDCR, GPSCR, BDCR, MECR, PFDR
        .db     PFCR, PFFR, PFDCR, PFDDR, PGDR, PGCR, PGFR, PGDCR
        .db     PGDDR, PWL0R, PWM0R, PWL1R, PWM1R, PWL2R, PWM2R, PWL3R
        .db     PWM3R, ICCSR, ICCR, ICT1R, ICT2R, ICS1R, ICS2R, QDCSR
        .db     QDCR, TAT1R, TACR, TAT2R, TAT8R, TAT3R, TAT9R, TAT10R
        .db     TAT5R, TAT6R, TAT7R, TBM1R, TBL1R, TBM2R, TBL2R, SECR
        .db     SAER, SFCR, SBER, SCER, SDER, SEER, SFER
;
..zTblSize=$-..zTbl
;
;  Running in RAM at this point.  The following code can
;  be executed but only in RAM debugging mode - be careful
;  what is placed here - it won't run in the Flashed version.
;
;  Initialize I/O and variables for running in RAM.
;  Shadows of I/O already initialized by the debug kernel are initialized.
;
initInRam:
        .if     INC_DEBUG
        ld      a,(_TAT4R)              ; Init srTAT4R shadow
        ld      (_srTAT4R),a
        ioi ld  (TAT4R),a               ; And port
        ld      a,(_GCDR)               ; Init srGCDR shadow
        ld      (__GCDR),a              ; Save for going in and out of doubling
        ld      (_srGCDR##),a
        ioi ld  (GCDR),a                ; Restore doubler as on startup
        ld      a,0C0h                  ; Init srPCSR
        ld      (_srPCFR##),a
        ld      hl,14h                  ; Debug kernel left them here
        ld      de,_srMB0CR##           ; Set MBxCR shadows
        ld      bc,4
        ldir
        ld      a,(16h)                 ; Current MB2CR
        and     3                       ; Set to CS0?
        ld      (__runInCS2),a          ; Will be 0 if mapped to flash (flash is CS0)
        ld      a,(_19200TC)
        ld      (__serial19200TC),a     ; Save 19200TC
        jr      nz,..no                 ; Don't add wait states in RAM
;
;  If the 19200 TC indicates we need wait states, one is added, even
;  in RAM to keep the timing between RAM and flash the same.
;
        cp      a,52                    ; 0 wait state cutoff (30MHz)
        jp      c,..no
        bool    hl                      ; 1 wait state (can be 2 or 4)
        push    hl
        ld      l,h                     ; 0 means set all MBxCRs
        push    hl
        lcall   $setWaitStates##
        add     sp,4
..no:
;
;  The debug kernel could have enabled doubling.  Turn it off if on.
;
        .ifdef  DISABLE_CLOCK_DOUBLER
        bool    hl                      ; Disable
        ld      l,h
        lcall   $enableClockDoubler##
        .endif
;
        ipset   2                       ; So that serial break-outs work
        xor     a,a                     ; A NOP
        ld      (10h),a                 ; In RAM, remove RETs to allow printfs
        ld      (18h),a                 ; Remove this code to stub printf in RAM
        ld      (39h),a
        call    3Dh                     ; Clear kb input buffer
        .endif
;
        ret
;
;  Calculate 19200 TC.  This is required for setting baud rates and
;  to see if clock doubling can be enabled.
;  If SET_19200TC is defined to a value, the runtime speed check to get
;  the 19200 time constant is skipped and the defined value is used.
;
initInFlash:
        .ifndef SET_19200TC
        call    SerialCalc19200TC
        .else
        ld      a,SET_19200TC
        .endif
;
        cp      a,36                    ; > 44MHz?
        jr      c,..noCpy               ; Skip if not
        ex      af,af'
;
;  At 44MHz, copy flash to 0 wait state RAM
;
        .ifndef RAM_COMPILE
        ld      hl,0C6C6h               ; 0WS CS 2
        ioi ld  (MB2CR),hl              ; Map bank 2 and 3
        ld      bc,8                    ; XPC 0 and 8
        ld      ix,0
        ld      de,2                    ; 32k*2=64k blocks
..cpy:  ld      a,b
        ldp     hl,(ix)
        ld      a,c
        ldp     (ix),hl
        add     ix,de
        jr      nc,..cpy
        inc     b
        inc     c
        bit     3,b                     ; B == 8?
        jr      z,..cpy                 ; Copy 512k
        ld      hl,(MBXCRvals+2)        ; Map MB2CR/MB3CR back
        ioi ld  (MB2CR),hl
        .endif
        ld      hl,0C6C6h               ; 0WS CS 2 (can use CECEh to write protect RAM)
        ioi ld  (MB0CR),hl              ; Map bank 0 and 1
;
;  These lines keep RAM addresses as linked (not unrelocated by the MMU as is
;  done for the normal flash initialization.  This keeps the stack and
;  variables in fast RAM (but it's not battery backed).  FARBSS is also in fast
;  RAM and can overflow into slow RAM.  In this case, just let FARBSS follow
;  STACK when linking for RAM debug and for flash.
;
        pop     de                      ; Get ret addr - change below moves memory
;;NSCL        ld      a,(sgat BSS >> 12) - (sgst BSS >> 12)
        ld      a,(sgat STACK >> 12) - (sgst STACK >> 12)  ;;NSCL
        ioi ld  (STACKSEG),a
        push    de
        push    hl
        call    initCram                ; Initialize BSS and initialized data
        pop     hl
        ld      (_srMB0CR##),hl         ; Update shadows
        ld      hl,(MBXCRvals+2)
        ld      (_srMB2CR),hl           ; Update shadows
        ld      a,1
        ld      (__runInCS2),a          ; Set flag in CS2
        ex      af,af'
        ld      (__serial19200TC),a     ; Save TC
        jr      ..512
;
..noCpy:ex      af,af'
        call    initCram                ; Initialize BSS and initialized data
        ex      af,af'
        ld      (__serial19200TC),a     ; Save TC
        ld      hl,MBXCRvals            ; Initialize MBxCR shadows
        ld      de,_srMB0CR##
        ld      bc,4
        ldir
        .ifndef RAM_COMPILE
        ld      hl,(07Bh)               ; ID Block Flash 1 size in k
        ld      de,512                  ; Check for 512k
        or      a,a
        sbc     hl,de                   ; ZF=1 if 512k
        jr      z,..512                 ; Skip if 512k
        ld      hl,_srMB1CR##
        ld      e,MB1CR
        ld      (hl),MB1CRXval          ; 256k MB1CR value
        ioi ldi
        .endif
..512:
;
;  Enable clock doubling if possible.  This could enable 1 wait state
;  on all chip selects if the speed is above 30MHz
;  For low power operation, define DISABLE_CLOCK_DOUBLER or do so later
;  in main using enableClockDoubler( 0 );
;
        .ifndef DISABLE_CLOCK_DOUBLER
        ld      hl,1                    ; Enable
        lcall   $enableClockDoubler##
        .endif
;
        ret
;
;  Zero out the BSS data area.
;
initCram:
        ld      a,(_resetCode)  ;;NSCL - save resetCode value
        push    af              ;;NSCL
        ld      hl, sgst BSS            ; Logical start of BSS
        ld      bc, sgsz BSS            ; Size of BSS
..zero: ld      a,b
        or      c
        jr      z,..restoreCode ;;NSCL - jump to restore point
        ld      (hl),0                  ; 0 first byte
        inc     hl
        dec     bc
        jr      ..zero
..restoreCode:
        pop     af              ;;NSCL - restore resetCode value
        ld      (_resetCode),a  ;;NSCL

;
;  Zero out the BSSZ data area.  This is for data with initializers of
;  0 (e.g. int x = 0;)
;
..bssz: ld      hl, sgst BSSZ           ; Start of BSSZ
        ld      bc, sgsz BSSZ           ; Size of BSSZ
..loop: ld      a,b
        or      c
        jr      z,..data
        ld      (hl),0                  ; Zero BSSZ
        inc     hl
        dec     bc
        jr      ..loop
;
;  Copy initialized data to segment IDATA.
;
..data: ld      iy, word (sgat DATA)    ; Phys start of DATA
        ld      e, sgat DATA >> 16
        ld      ix, sgst IDATA          ; Logical start of IDATA
        ld      bc, sgsz DATA           ; Size of DATA
..copy: ld      a,b
        or      c
        ret     z                       ; return if IDATA size is zero
        dec     bc
        ld      a,e                     ; No data
        ldp     hl, (iy)
        ld      (ix), l                 ; Copy
        inc     ix
        inc     iy
        ld      hl,iy
        bool    hl
        jr      nz,..copy
        inc     e
        jr      ..copy
;
;  Initialize xmem starting address and size.  This can be changed
;  if you want a fixed pool of a fixed size.  It's implemented here
;  with a simple memory sizing routine to maximize xmem in RAM and Flash.
;  This must do some work using the values in the MMU to find the real
;  address of memory since it varies in a RAM and flash enviroment.
;
$initXmem::
        push    ix
        ld      ix,__xmemStart

;;        .if 0
;
;  Calculate begining of xmem.  This is the end of STACK or FARBSS if it exists,
;  but in Flash, if FARBSS is empty, BSS is moved by the MMU, so this follows
;  the move in this case.  If using FARBSS, STACK has to be moved linking for
;  flash and FARBSSes end address is ok.
;
;        ld      a,SGSZ FARBSS != 0      ; Runtime get FARBSS size != 0
;        or      a,a
;        jr      nz,..useFARBSS          ; Use end of FARBSS if it's there
;        ld      a,(inFlash)             ; If not in flash, also use FARBSS end
;        or      a,a
;        jr      z,..useFARBSS
;;NSCL        ld      a,((SGAT STACK + SGSZ STACK + 0xFFF - SGAT BSS) >> 16) + 8
;;NSCL        ld      hl,word (SGAT STACK + SGSZ STACK + 0xFFF - SGAT BSS) & 0xF000
;        ld      a,(sgat STACK + sgsz STACK) >> 16  ;;NSCL
;        ld      hl,word (sgat STACK + sgsz STACK)  ;;NSCL
;        jr      ..cont
;..useFARBSS:
;        ; Done this way in case FARBSS is missing
;        ld      a,(sgat FARBSS + sgsz FARBSS) >> 16
;        ld      hl,word (sgat FARBSS + sgsz FARBSS)
;;       ld      a,(sgat STACK + sgsz STACK + sgsz FARBSS) >> 16
;;       ld      hl,word (sgat STACK + sgsz STACK + sgsz FARBSS)

;        .else

;   ;; hl = (__stack + 0xfff) >> 4;
;        ld      hl,__stack
;        ld      de,0xfff
;        add     hl,de
;        rr      hl
;        rr      hl
;        rr      hl
;        rr      hl
;    ;; a = ((hl >> 4) & 0x0f) + (STACKSEG);
;        ld      a,h
;        and     0xf
;        ld      iy,STACKSEG
;        ioi add a,(iy)
;        ; a=stack phys/4096
;        ; a:hl is stack
;
;        bool    hl
;        ld      l,h
;        ld      h,a
;        xor     a
;        add     hl,hl
;        rla
;        add     hl,hl
;        rla
;        add     hl,hl
;        rla
;        add     hl,hl
;        rla
;        ; a=64k seg, HL=4k seg
;
;        ld      b,SGSZ FARBSS != 0      ; Runtime get FARBSS size != 0
;        bit     0,b
;        jr      z,..cont
;;..useFARBSS:
;        ; Done this way in case FARBSS is missing
;        ld      de,word sgsz FARBSS
;        add     hl,de
;        jr      nc,..same_64k
;        inc     a
;..same_64k:
;        ld      b,sgsz FARBSS>>16
;        add     b
;
;        .endif
;
;..cont:
;;        ld      a,(sgat XMEMSTART + sgsz XMEMSTART) >> 16  ;;NSCL
;;        ld      hl,word (sgat XMEMSTART + sgsz XMEMSTART)  ;;NSCL
        ld      a,(sgat XMEMSTART) >> 16  ;;NSCL
        ld      hl,word (sgat XMEMSTART)  ;;NSCL
;
;  AHL is the start of xmem
;
        ld      (ix),hl                 ; Store LSW in variable
        ld      (ix+8),hl               ; And heap pointer too
        ex      de,hl
        bool    hl
        ld      l,a
        ld      (ix+2),hl               ; Set MSW
        ld      (ix+10),hl              ; Set heap pointer too
        ex      de,hl
        jr      ..ram
;
;  Find end of memory - AHL is the start of xmem.
;
..safe: .ds     2                       ; A safe place in case we hit this code
..ram:  ld      de,1                    ; Remove 1 to get end of STACK
        sbc     hl,de
        jr      nc,..nb                 ; Skip if no borrow
        dec     a
..nb:   ld      c,a                     ; CIY is xmem base
        ld      a,0F0h
        and     a,h
        or      a,(..safe >> 8) & 0x0F
        ld      h,a
        ld      l,..safe & 0FFh         ; Ensure CIY won't hit this code
        ld      iy,hl
        ld      a,c                     ; CIY is the end of the first 64k block of memory
        ldp     hl',(iy)                ; Get word
        ld      hl,5A5Ah
        ldp     (iy),hl                 ; Mark first block
        ioi ld  hl,(RTC0R)              ; Load RTC val for a "random" test word
        ld      (ix+4),hl               ; Hold here for test
..lp:   inc     a                       ; Next 64k
        cp      a,12h                   ; 1FFFFF is too far, end
        jr      z,..end
        ld      b,a
        ldp     hl,(iy)                 ; Get word
        ex      de,hl
        ld      hl,(ix+4)
        ldp     (iy),hl
        ldp     hl,(iy)
        ld      a,l
        cp      a,(ix+4)
        jr      nz,..end                ; No RAM, done
        ld      a,h
        cp      a,(ix+5)
        jr      nz,..end
        bit     0,b                     ; Continue on odd 64k pages
        jr      nz,..4
        ld      a,b                     ; Do 128 mirror checks
..5:    sub     a,2                     ; Move down looking for mirror
        jr      c,..4                   ; Not past 0
        ldp     hl,(iy)
        ex      af,af'
        ld      a,l
        cp      a,(ix+4)
        jr      nz,..6                  ; No mirror
        ld      a,h
        cp      a,(ix+5)
        jr      z,..7                   ; Mirror to lower memory, stop
..6:    ex      af,af'
        jr      ..5
..7:    ex      de,hl
        ld      a,b
        ldp     (iy),hl                 ; Restore mirror word
        jr      ..end
..4:    ld      a,c
        ldp     hl,(iy)                 ; Hit first word?
        ld      a,b
        ex      de,hl
        ldp     (iy),hl                 ; Restore word
        ld      a,5Ah
        cp      a,e                     ; First word unchanged?
        jr      nz,..end                ; No, done
        cp      a,d
        ld      a,b
        jr      z,..lp
..end:  ld      a,c
        exx
        ldp     (iy),hl                 ; Restore first word
        exx
        bool    hl
        ld      l,b
        ld      (ix+14),hl
        ld      l,h
        ld      (ix+12),hl
        ex      de,hl
        ld      hl,(ix)                 ; Low word of base
        ex      de,hl
        sbc     hl,de
        ld      (ix+4),hl               ; Size LSW
        ld      hl,(ix+14)
        ex      de,hl
        ld      hl,(ix+2)
        inc     hl                      ; Next page
        sbc     hl,de
        ld      (ix+6),hl               ; Size MSW
        pop     ix
        lret
;
;  Calculate the timer constant for 19200.
;  Returned in A.
;  This must be called with 0 wait states!
;
        .ifndef SET_19200TC
SerialCalc19200TC:
        ld      a,5Ah
        ld      de,1FFFh        ; mask for RTC bits
..wait: ioi ld  (RTC0R),a       ; load RTC registers
        ioi ld  hl,(RTC0R)      ; get LSW RTC
        and     hl,de           ; mask off bits
        jr      nz,..wait       ; wait bits 0-9 = 0
        ex      de,hl           ; DE = count = 0
..loop: inc     de
        ld      b,49
        djnz    $
        ioi ld  (WDTCR),a
        ioi ld  (RTC0R),a
        ioi ld  hl,(RTC0R)      ; get LSW RTC
        bit     4,h             ; test bit 10
        jr      z,..loop
        ld      a,d             ; 19200 time constant
        ret
        .endif
;
;  Get function pointers in segment STARTUP and call each one.
;  This allows compile-time installment of startup initialization
;  which can be executed before main.  The order of call is the
;  order of the addresses in the output.  In a single module they
;  are guaranteed to occur in order.  Modules linked ahead of others
;  are guaranteed to execute their STARTUP functions first.
;  Included only if STARTUP_SEG is defined because not many users use
;  this and therefore the code is wasted.
;
        .ifdef  STARTUP_SEG
statupFunctions:
        ld      a,sgsz STARTUP / 2      ; Number of entries
        or      a
        jr      z,..main
        ld      hl,sgst STARTUP
..glp:  ld      e,(hl)
        inc     hl
        ld      d,(hl)
        inc     hl
        push    hl
        push    af
        call    __callde##
        pop     af
        pop     hl
        dec     a
        jr      nz,..glp
        ret
        .endif
;
end:    .debug  } end 827
;
;  Define variables defining free RAM available for the heap.
;  If the heap is moved, __heap_end must remain >= __heap_start.
;  If malloc/free are not used, then it doesn't matter.  These
;  must address memory that is always mapped in.  It's probable
;  but very important that the heap start is NOT 0.
;
;  We highly recommend against using malloc and free in an
;  embedded application.
;
;  Te define a head, set HEAP_SIZE in the Asm defines for cstart
;  or change the HEAP_SIZE definition below only when HEAP_SIZE
;  isn't already defined.
;
        .ifndef HEAP_SIZE
HEAP_SIZE = 0
        .endif
;
__heap_start::
        dw      _heap
__heap_end::
        dw      _heap_end
;
        .bseg
_heap:  .ds     HEAP_SIZE
_heap_end:
;
        .bseg
;
;  Calculated time constant for TAT4R.  Use by serial libraries
;  to handle baud rates without regard to CPU speed.  If GCDR
;  is changed, this value must also be updated (going into
;  double clock mode means you must double this constant!)
;  This must be adjusted accordingly if GCDR is changed!
;
__serial19200TC::
        .ds     1
;
;  Reason for reset is stored here:
;    00 - Program restarted at 0
;    01 - Watchdog caused reset
;    03 - Reset line caused reset
;
_resetCode::
        .ds     1
;
;  Byte written to GCDR.  Can be used to go in and out of double speed.
;
__GCDR::
        .ds     1
;
;  Running in 0 wait state RAM on CS2 for 44MHz modules.
;
__runInCS2::
        .ds     1
;
;  Define xmemory base, size, and heap pointer.  The xmem heap is sized at runtime
;  to allow the maximum xmem space in RAM and flash.  These must be kept together
;  and as defined.
;
__xmemStart::
        .ds     4
__xmemSize::
        .ds     4
__xmemHeap::
        .ds     4
__xmemEnd::
        .ds     4
;
;  An initialized FILE structure for initial printf redirection
;
        .dseg
__debugStdio:                           ; This is a struct FILE
        .dw     18h                     ; Output address
        .dw     10h                     ; Input address
        .dw     0                       ; Putback char
        .dw     0, 0, 0
;
        .if     INC_DEBUG
start   =       cstart
        .else
start   =       0xFFFF                  ; FFFF means no start (stops autorun)
        .endif
        .end    start
