//----- ps8Rack.c ----- 10/24/12 -----(03/26/06)-----

// EE Documents bundle # 038

//-----------------------------------------------------------------------------
//  Code to handle hardware-specific implementation of generic hdw functions
//-----------------------------------------------------------------------------
//  Change Log (most recent first):
//
//
//  2012-10-24 - Mark Davis
//               - Changed all the DC-style WrPortI() func calls to the more
//                 efficient Softools outi() macro.
//
//  2010-02-06 - Mark Davis
//               - Fixes to convertADS7825(), ReadDigInputs(),
//                 WriteDigOutputs(), & hdwInit() to avoid race conditions
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>

#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>

#include <sys/farprintf.h>

//--- see hdwInit() for Digital I/O pin usage ---

const char ArchName[] = { "PS8Rack" };

#define  NUM_ADC   16
#define  NUM_DAC    8

#define  MAX_DIN   32
#define  MAX_DOUT  32


//--- these values can be set LOWER by the application ---
const uint  NumHdwADC = NUM_ADC;
const uint  NumHdwDAC = NUM_DAC;

const uint  MaxDIN =  MAX_DIN;
const uint  MaxDOUT = MAX_DOUT;


//--- NSCL 16-bit analog RCM3200-based ctlrs use port B for SPI to ADCs, DACs
#define  ADC_DAC_SPIPort  1  // for most RCM3200-based NSCL ctlr's w/Analog


unsigned char  __brdInitFlag;

#include <hdwDefs.c>


SerialInfo  *spiPort;


//=============================================================================
_nearcall float adcGetRaw(uint ui)  { return (float) adcRaw[ui]; }

_nearcall float adcMinRaw(uint ui)  { return 0.0; }
_nearcall float adcMaxRaw(uint ui)  { return (float)0xffff; }

_nearcall void adcDefaults(uint ui, float *gain, float *offset)
{ *gain = 0.0003046;  *offset = 32786.0; }

_nearcall void adcCalibPoints(uint ui, float *low, float *high)
{ *low = -9.5;  *high = 9.5; }

_nearcall char *adcUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================
_nearcall float dacGetRaw(uint ui)  { return (float) dacRaw[ui]; }

_nearcall float dacMinRaw(uint ui)  { return 0.0; }
_nearcall float dacMaxRaw(uint ui)  { return (float)0xffffL; }

_nearcall void dacDefaults(uint ui, float *gain, float *offset)
{  *gain = 0.000311;  *offset = 32768.0;  }

_nearcall void dacCalibPoints(uint ui, float *low, float *high)
  {  *low = -9.5;  *high = 9.5; }

_nearcall char *dacUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
Function: int _readExtioPort ( int port );

Note:
This function is setup to provide additional hold time when
accessing external I/O addresses 0000 and 0002. This routine
must stay         and not be changed, for the additional
address hold-time to be valid.

Improved by Softools to not require a copy to RAM because
real tools have align and other stuff for real programmers.
************************************************************/
/*
unsigned readExtIO (unsigned port)      // Port is 0 or 2
{
#asm
    push    ip
    ipset   3
    ld      a,l
    and     a,0x03
    jr      z,..read0
    jr      ..skip
    .align  4           ; Align to clear A0 & A1
..read0:
    nop                 ; A0 = 0, A1 = 0
    nop                 ; A0 = 1, A1 = 0
    ioe     ld l,(hl)   ; A0 = 0, A1 = 1
    pop     ip          ; A0 = 0, A1 = 0
    jr      ..exit
    .align  4           ; Align to clear A0 & A1
..skip:
    ioe     ld l,(hl)   ; A0 = 0, A1 = 0 (2 bytes)
    pop     ip          ; A0 = 0, A1 = 1
..exit:
#endasm
}
*/

//-----------------------------------------------------------------------------
//  determine the device ID based on external hardware settings
//-----------------------------------------------------------------------------
ulong ReadDeviceID()
{
  uchar  inA, devID;


  inA = ini(PADR);

//--- device ID comes from 8-bit switch input ----
  devID = ((inA & 0x01) << 7)
        | ((inA & 0x02) << 5)
        | ((inA & 0x04) << 3)
        | ((inA & 0x08) << 1)
        | ((inA & 0x10) >> 1)
        | ((inA & 0x20) >> 3)
        | ((inA & 0x40) >> 5)
        | ((inA & 0x80) >> 7);
  devID = ~devID;

  return (ulong) devID;
}

//-----------------------------------------------------------------------------
//  Return the current state of the digital inputs
//
//  The digital inputs are connected to a series of octal line drivers whose
//  outputs are connected to the parallel G signal lines.  To read all the
//  inputs, we enable the drivers one at a time and read the state of the G
//  lines, adding the result to previously read bits.
//
//  The digital inputs and outputs are address in groups of 8 bits.  Which
//  group of 8 bits we are setting or reading is controlled by a 3-to-8 line
//  decoder, where toggling one of its 8 outputs low latches the desired
//  data.
//
//  To avoid activating the wrong one of the 8 outputs even for a moment, we
//  have to make sure the 3 inputs that specify which output to toggle low
//  are all stable BEFORE we enable the outputs and while we are disabling
//  the outputs.
//-----------------------------------------------------------------------------
ulong ReadDigInputs()
{
  ulong  inBits;
  uchar  otherBits = srPBDR & 0x87;


//--- To insure that only the intended Y output changes state, we need to ---
//--- apply the bits that select which output, enable the Y outputs, ---
//--- then disable the Y outputs without changing the selection bits. ---
//--- If we change the selection bits at the same time that we enable ---
//--- the Y outputs, then we run the risk of causing other outputs to ---
//--- briefly change state.  This could be especially bad if one of them ---
//--- causes one of the sets of digital output latches to change state ---

//--- read input bits 24-31 --- (Y7)
  outi(PBDR, srPBDR = otherBits | 0x78 );     // select /Y7
  outi(PBDR, srPBDR = otherBits | 0x58);      // /Y7 low to enable driver
  inBits = (ulong) ini(PGDR);                 // read the data
  outi(PBDR, srPBDR = otherBits | 0x78);      // /Y7 high again

//--- read input bits 16-23 --- (Y6)
  outi(PBDR, srPBDR = otherBits | 0x38 );     // select /Y6
  outi(PBDR, srPBDR = otherBits | 0x18 );     // /Y6 low to enable driver
  inBits = (inBits << 8) | (ulong)ini(PGDR);  // read the data
  outi(PBDR, srPBDR = otherBits | 0x38 );     // /Y6 high again

//--- read input bits 08-15 --- (Y5)
  outi(PBDR, srPBDR = otherBits | 0x70 );     // select /Y5
  outi(PBDR, srPBDR = otherBits | 0x50 );     // /Y5 low to enable driver
  inBits = (inBits << 8) | (ulong)ini(PGDR);  // read the data
  outi(PBDR, srPBDR = otherBits | 0x70 );     // /Y5 high again

//--- read input bits 00-07 --- (Y4)
  outi(PBDR, srPBDR = otherBits | 0x30 );     // select /Y4
  outi(PBDR, srPBDR = otherBits | 0x10 );     // /Y4 low to enable driver
  inBits = (inBits << 8) | (ulong)ini(PGDR);  // read the data
  outi(PBDR, srPBDR = otherBits | 0x70 );     // /Y4 high again


  outi(PBDR, srPBDR = otherBits | 0x20);  // deselect all drivers

  inBits = (inBits & ~forcedDigInMask) | forcedDigInBits;

  return inBits;
}


//-----------------------------------------------------------------------------
//  set the state of the digital outputs
//
//  The digital inputs and outputs are address in groups of 8 bits.  Which
//  group of 8 bits we are setting or reading is controlled by a 3-to-8 line
//  decoder, where toggling one of its 8 outputs low latches the desired
//  data.
//
//  To avoid activating the wrong one of the 8 outputs even for a moment, we
//  have to make sure the 3 inputs that specify which output to toggle low
//  are all stable BEFORE we enable the outputs and while we are disabling
//  the outputs.
//-----------------------------------------------------------------------------
void WriteDigOutputs(ulong digOut)
{
  ulong  outBits;
  uchar  otherBits = srPBDR & 0x87;


  digOut = (digOut & ~forcedDigOutMask) | forcedDigOutBits;

  outBits = ~digOut;

//--- set digital I/O bus (port G) to output mode ---
  outi(PGDDR, srPGDDR = 0xFF);

//--- update output bits 00-07 --- (Y0)
  outi(PGDR, srPGDR = (uchar)outBits);          // put value on bus
  outi(PBDR, srPBDR = otherBits | 0x20);        // select /Y0
  outi(PBDR, srPBDR = otherBits );              // enable outputs (/Y0 low)
  outi(PBDR, srPBDR = otherBits | 0x20);        // disable outputs (all high)

//--- update output bits 08-15 --- (Y1)
  outi(PGDR, srPGDR = (uchar)(outBits >> 8));   // put value on bus
  outi(PBDR, srPBDR = otherBits | 0x60);        // select /Y1
  outi(PBDR, srPBDR = otherBits | 0x40 );       // enable outputs (/Y1 low)
  outi(PBDR, srPBDR = otherBits | 0x60);        // disable outputs (all high)

//--- update output bits 16-23 --- (Y2)
  outi(PGDR, srPGDR = (uchar)(outBits >> 16));  // put value on bus
  outi(PBDR, srPBDR = otherBits | 0x28);        // select /Y2
  outi(PBDR, srPBDR = otherBits | 0x08 );       // enable outputs (/Y2 low)
  outi(PBDR, srPBDR = otherBits | 0x28);        // disable outputs (all high)

//--- update output bits 24-31 --- (Y3)
  outi(PGDR, srPGDR = (uchar)(outBits >> 24));  // put value on bus
  outi(PBDR, srPBDR = otherBits | 0x68);        // select /Y3
  outi(PBDR, srPBDR = otherBits | 0x48 );       // enable outputs (/Y3 low)
  outi(PBDR, srPBDR = otherBits | 0x68);        // disable outputs (all high)

//--- set digital I/O bus (port G) back to input mode ---
  outi(PGDDR, srPGDDR = 0x00);
}


//-----------------------------------------------------------------------------
/*
void resetADC(void)
{
  uchar  buf[10];


  if (! spiPort)  return;

//--- set SPI clock mode ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_FALL);

//--- enable conversion mode ---
  outi(PFDR, srPFDR &= 0xF7);  // R/C- (PF3) low (convert)

//--- set CS- low to start the conversion ---
  outi(PFDR, srPFDR &= 0xfb);  // CS- (PF2) low

//--- an extra-long delay to insure the conversion completes ---
  for (uint xx=0; xx<100; ++xx);

//--- set R/C- high to enable read mode ---
  outi(PFDR, srPFDR |= 0x08);  // R/C- (PF3) high

//--- read whatever data it has ---
  _spiRead_(spiPort, buf, 9, false);

//--- then reset /CS to finish the cycle ---
  outi(PFDR, srPFDR |= 0x04);  // CS- (PF2) high

//--- and another delay for good measure ---
  for (uint xx=0; xx<100; ++xx);
}
*/

//-----------------------------------------------------------------------------
//  Function to communicate with an ADS7825 chip (a 4-channel, 16-bit ADC)
//
//  Call:
//    convertADS7825(uint nextChan)
//
//    Input:
//       nextChan: Channel # to sample between calls and to do conversion
//                 on during the next call.
//
//    Result:
//       The raw value for the channel specified during the previous call
//       is updated.
//
//
//  To maximize the conversion rate, the ADS7825 begins acquiring/sampling
//  the signal to be converted as soon as the previous conversion is
//  completed.  This requires that it already know what channel to begin
//  sampling, so you have to tell it what channel to sample for the NEXT
//  conversion, NOT the current conversion.
//
//  NOTE that this does NOT mean that the conversion initiated by this call
//  is for outdated data.  The ADS7825 is sampling the channel specified
//  during the last call right up to the time it is told to start the
//  conversion.  The only catch is you have to know what channel that was
//  so when you read the conversion data you know what channel it is for.
//
//  The ADC7825 is fairly unique in that it has a TAG input that allows
//  multiple ADC's to be connected in a daisy-chain configuration, allowing
//  you to initiate a conversion on multiple chips at the same time and then
//  read the results from all of them in a single operation.  This greatly
//  decreases the additonal time needed to support more than 4 channels when
//  the chips are connected in this manner.
//
//  NOTE:  The downside, of course, is the cost per input, both in $ and in
//         board space.  Only 4 inputs means never more than 4 passes to
//         read any # of channels.  It also means one more (expensive) chip
//         for each additional group of 4 channels.
//
//-----------------------------------------------------------------------------
void convertADS7825(uint nextChan)
{
  uchar  buf[10];
  static uint  prevChan = 0;


  nextChan &= 3;

//--- Set the bits that will tell the ADC which channel to sample between ---
//--- calls to this func, and to hold and convert during the next call ---
//--- (PE6 and PE7 are connected to ADR0 and ADR1 on all ADC's) ---
  outi(PEDR, srPEDR = (srPEDR & 0x3F) | (nextChan << 6));

//--- set R/C- low first so we can use CS- to start conversion ---
  outi(PFDR, srPFDR &= 0xF7);  // R/C- (PF3) low (convert)

//--- Toggle /CS low then high very quickly.  This starts the conversion ---
//--- on the previously specified input and latches the input # to be used ---
//--- for the next conversion.  /CS must be low for at least 40ns, but must ---
//--- go high again before BUSY does (aprox 25us).  If /CS is still low when ---
//--- BUSY goes high, the next conversion will start and the results of the ---
//--- previous one will be lost ---
  _EnterCritical_();
  outi(PFDR, srPFDR &= 0xFB);  // /CS (PF2) low
  outi(PFDR, srPFDR |= 0x04);  // /CS (PF2) high
  _ExitCritical_();

//--- set R/C- high to enable read mode ---
  outi(PFDR, srPFDR |= 0x08);  // R/C- (PF3) high

//--- The design did not include the ability to read the BUSY signal, so ---
//--- add a short delay to avoid trying to read the data before it is ready ---
  for (uint xx=0; xx<20; ++xx);

//--- set /CS low to enable transmission of data ---
  outi(PFDR, srPFDR &= 0xFB);  // /CS (PF2) low

//--- The ps8rack controller has 4 ADC's chained together, so we will get ---
//--- values for 4 channels (the same one from each ADC) at once.  ---
//--- NOTE: Each ADC sends 17 bits (for a total of 68 bits, or 8.5 bytes). ---
//---       The first bit of each 17 is NOT part of the data ---
  _spiRead_(spiPort, buf, 9, false);

//--- set /CS high so ADC does not react to clock pulses not meant for it ---
  outi(PFDR, srPFDR |= 0x04);  // /CS (PF2) high

//--- extract the raw value from the bytes just read ---
  adcRaw[prevChan] = (uint) (0x8000 ^ (
      ((uint)buf[0] << 9) |
      ((uint)buf[1] << 1) |
      ((uint)buf[2] >> 7)
    ) );
  adcRaw[prevChan+4] = (uint) (0x8000 ^ (
    ((uint)buf[2] << 10) |
    ((uint)buf[3] << 2) |
    ((uint)buf[4] >> 6)
  ) );
  adcRaw[prevChan+8] = (uint) (0x8000 ^ (
    ((uint)buf[4] << 11) |
    ((uint)buf[5] << 3) |
    ((uint)buf[6] >> 5)
  ) );
  adcRaw[prevChan+12] = (uint) (0x8000 ^ (
    ((uint)buf[6] << 12) |
    ((uint)buf[7] << 4) |
    ((uint)buf[8] >> 4)
  ) );

//--- keep track of what channel we started the conversion on, because that ---
//--- is the channel whose value will be read during the next call ---
  prevChan = nextChan;
}


//-----------------------------------------------------------------------------
//  Read analog inputs.  The PS8Rack controller has 4 ADS7825 ADC's.  Each
//  ADC has 4 16-bit inputs and the 4 chips are daisy-chained together so
//  data can be read from all 4 chips in a single operation.
//-----------------------------------------------------------------------------
void readAnalogInputs(ulong mask)
{
  if (! spiPort)  return;

//--- Remember we have to tell the ADC which channel's data we will read ---
//--- NEXT time we call it, so consistently using the sequence 1, 2, 3, 0 ---
//--- each time means we get the data for 0, 1, 2, 3.  Also remember that ---
//--- this means channel 0 will be continously sampled (READY for conversion) ---
//--- between calls to this function ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_FALL);
  convertADS7825(1);  // chan to convert NEXT call
  convertADS7825(2);  // chan to convert NEXT call
  convertADS7825(3);  // chan to convert NEXT call
  convertADS7825(0);  // chan to convert NEXT call
}


//-----------------------------------------------------------------------------
//  Update the state of the AD5542 DAC chips
//-----------------------------------------------------------------------------
void writeAnalogOutputs(ulong mask)
{
  uchar  buf[2];
  uint   rawVal, bitTest;
  uchar  i;


  if (! spiPort)  return;

//--- set SPI clock mode ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_RISE);

  for (i=0,bitTest=1; i<NUM_DAC; ++i,bitTest<<=1)  if (bitTest & (uint)mask)  {

  //--- select one of the 8 DAC chips ---
  //--- (chip # set by PD5-PD4-PE3) ---
    outi(PEDR, srPEDR = (srPEDR & 0xF7) | (i&1)<<3 );
    outi(PDDR, srPDDR = (srPDDR & 0xCF) | (i&6)<<3 );

  //--- enable selected decoder output line to set /CS low for one DAC ---
    outi(PBDR, srPBDR &= 0x7F);  // PB7 ctls decoder output

  //--- send cmd to DAC chip ---
    rawVal = dacRaw[i];   buf[0] = rawVal >> 8;  buf[1] = rawVal & 0xff;
    _spiWrite_(spiPort, buf, 2, false);

  //--- disable decoder outputs to bring /CS high for all DAC's ---
    outi(PBDR, srPBDR |= 0x80);  // PB7 ctls decoder output
  }

}


//-----------------------------------------------------------------------------
/*
float ADCtoVolts(uchar chan, uint rawADC)
{
  return (ADCOffset(chan) - (float)(rawADC)) * ADCGain(chan);
}
*/

//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to transmit mode
//-----------------------------------------------------------------------------
void setExt485Tx(void)
{
}
//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to receive mode
//-----------------------------------------------------------------------------
void setExt485Rx(void)
{
}

//-----------------------------------------------------------------------------
//  functions to enable/disable the transmitter for an RS-485 connection
//-----------------------------------------------------------------------------
void _nearcall SerSetDirA(int enableXmit)  {  }
void _nearcall SerSetDirB(int enableXmit)  {  }
void _nearcall SerSetDirC(int enableXmit)  {  }
void _nearcall SerSetDirD(int enableXmit)
{  // PF5 controls RS-485 line driver for port D(3) RS-485 connections
  char  val;
  val = enableXmit ? (srPFDR | 0x20) : (srPFDR & ~0x20);
  outi(PFDR, srPFDR = val);
}


//-----------------------------------------------------------------------------
void hdwInit(ulong digOutState)
{
  hdwType = HT_PS8_RACK;

  __brdInitFlag = TRUE;

// (NI) = Normal Input
// (NO) = Normal Output
// (AI) = Alternate Input
// (AO) = Alternate Output
// (xx) = Used by ZWorld core - do not change

  //--- Port A --- (*must be inputs due to bug in earlier Rabbit 3000's)
  //  PA7 - (NI) *
  //  PA6 - (NI) *
  //  PA5 - (NI) *
  //  PA4 - (NI) *
  //  PA3 - (NI) *
  //  PA2 - (NI) *
  //  PA1 - (NI) *
  //  PA0 - (NI) *
    outi(SPCR, srSPCR &= ~0x0c);  // PortA: byte input

  //--- Port B ---
  //  PB7 - (NO)  G2A on 3-to-8 decoder for Analog output selector
  //  PB6 - (NO)  A0 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB5 - (NO)  /OE2A input on 3-to-8 line decoder for Dig I/O Bus
  //  PB4 - (NO)  A2 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB3 - (NO)  A1 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB2 - (NO)  /OE input on all Digital output line drivers
  //  PB1 - (AO)  Serial-A (Diag/Prog) clock
  //  PB0 - (AO)  Serial-B (SPI) clock (connected to all ADCs and DACs)
    outi(PBDR, srPBDR |= 0xA0);   // PB7 and PB5 high
    outi(PBDDR, srPBDDR |= 0xFC); // dir: OOOO-OOxx

  //--- Port C ---
  //  PC7 - (xI)  Serial-A (Diag/Prog) receive
  //  PC6 - (AO)  Serial-A (Diag/Prog) transmit
  //  PC5 - (xI)  Serial-B (SPI) receive (from 1st ADC in the chain)
  //  PC4 - (AO)  Serial-B (SPI) transmit (to all DAC's)
  //  PC3 - (xI)  Serial-C (RS-232) receive
  //  PC2 - (AO)  Serial-C (RS-232) transmit
  //  PC1 - (xI)  Serial-D (RS-232/485) receive
  //  PC0 - (AO)  Serial-D (RS-232/485) transmit
    outi(PCFR, srPCFR = 0x55);  // func: xAxA-xAxA

  //--- Port D ---
  //  PD7 - (xx)  Ethernet Controller (TPIN+)
  //  PD6 - (xx)  Ethernet Controller (TPIN-)
  //  PD5 - (NO)  C input on 3-to-8 decoder for Analog output selector
  //  PD4 - (NO)  B input on 3-to-8 decoder for Analog output selector
  //  PD3 - (xx)  Ethernet Controller (TPOUT+)
  //  PD2 - (xx)  Ethernet Controller (TPOUT-)
  //  PD1 - (xx)  Not Avail (pulled high - not even connected???)
  //  PD0 - (xx)  Ethernet Controller Reset
    outi(PDDDR, srPDDDR |= 0x30);   // dir:  xxOO-xxxx
    outi(PDFR, srPDFR &= ~0x30);    // func: xxNN-xxxx

  //--- Port E ---
  //  PE7 - (NO)  ADR1 on all ADC's (output)
  //  PE6 - (NO)  ADR0 on all ADC's (output)
  //  PE5 - (NI)  unused (AI = Interrupt Input 1B (buffered from extern source)
  //  PE4 - (NI)  unused (AI = Interrupt Input 0B (buffered from extern source)
  //  PE3 - (NO)  A on 3-to-8 decoder for Analog output selector
  //  PE2 - (xx)  Ethernet Controller Chip Select (output)
  //  PE1 - (AI)  Interrupt Input 1A (buffered from external source)
  //  PE0 - (xx)  Interrupt Input 0A (Ethernet ctlr Interrupt)
    outi(PEDDR, srPEDDR = (srPEDDR | 0xC8) & ~0x32); // dir:  OOII-OxIx
    outi(PEFR, srPEFR = (srPEFR | 0x02) & ~0xF8);    // func: NNNN-NxAx

  //--- Port F ---
  //  PF7 - (AO)  PWM0 (buffered output to external device)
  //  PF6 - (AO)  PWM1 (buffered output to external device)
  //  PF5 - (NO)  xmit enable for Serial D (output)
  //  PF4 - (NO)  xmit enable for Serial C (output)
  //  PF3 - (NO)  to R/C on all ADC's (output)
  //  PF2 - (NO)  to CS on all ADC's (output)
  //  PF1 - (AO)  Serial-C clock (to RS-485/RS-232 port)
  //  PF0 - (AO)  Serial-D clock (to RS-485/RS-232 port)
    outi(PFDR, srPFDR |= 0x0C);   // PF2/3 high (ADC R/C-, /CS)
    outi(PFDDR, srPFDDR = 0xFF);  // dir:  OOOO-OOOO
    outi(PFFR, srPFFR = 0xC3);    // func: AANN-NNAA

  //--- Port G ---
  //  PG7 - (NIO)  Digital I/O bus
  //  PG6 - (NIO)  Digital I/O bus
  //  PG5 - (NIO)  Digital I/O bus
  //  PG4 - (NIO)  Digital I/O bus
  //  PG3 - (NIO)  Digital I/O bus
  //  PG2 - (NIO)  Digital I/O bus
  //  PG1 - (NIO)  Digital I/O bus
  //  PG0 - (NIO)  Digital I/O bus
    outi(PGDDR, srPGDDR = 0x00);  // dir:  IIII-IIII (for now)
    outi(PGFR, srPGFR = 0x00);    // func: NNNN-NNNN


  //--- all parallel port outputs driven hi/lo (no open-drain outputs) ---
    outi(PDDCR, srPDDCR = 0x00);
    outi(PFDCR, srPFDCR = 0x00);
    outi(PGDCR, srPGDCR = 0x00);

  //--- prep state of digital outputs ---
    digitalOut = digOutState;  WriteDigOutputs(digitalOut);

  //--- enable output latches so digital output states take effect ---
    outi(PBDR, srPBDR &= ~0x04);  // bring PB2 low

//    ser485Rx();  // disable the RS485 transmitter
//    setExt485Rx();  // set external 2/4-wire RS-485 adapter to Rx mode

    spiPort = NULL;  // set later by a call from the kernel process



  flashSize = 524288L;

//  resetADC();

  printf("\r\nInitialized %s hardware\r\n", ArchName);
}


//-----------------------------------------------------------------------------
//  Initialize and register hardware-dependent file systems
//-----------------------------------------------------------------------------
void registerFileSystems(void)
{
  socketInit_();
  srfsInit_();
  serialInit_();

  flashFileInit_();
}

//-----------------------------------------------------------------------------
//  Initialize the network hardware
//-----------------------------------------------------------------------------
void netHdwInit()
{
//  Select ethernet hardware:
//  usingAsix, usingDevKit, usingRealtek, usingSMSC
//  usingAll() can be used to support all cmd core modules, but req's more code

  usingAsix();
}


//-----------------------------------------------------------------------------
//  allocate any resources that should be owned by the kernel process
//
//  To be called ONLY by the kernel process!!!!
//-----------------------------------------------------------------------------
void hdwInit2(void)
{
  fd_t  spifd;


  spifd = _serial_(ADC_DAC_SPIPort, 0, SER_SPI, SER_NO_FLOW);

  spiPort = _FileFromFD_(spifd)->ioDev;

  printf("spiPort: %04lX (%d)\r\n", (ulong)spiPort, spiPort->port);  //tdebug
}

//-----------------------------------------------------------------------------

