//----- DLMIII.h ----- 03/16/09 -----(xx/xx/06)-----

/*-----------------------------------------------------------------------------
  SHDesigns DLMIII.h file

  This file defines structures used by the SHDesigns flash file system.

  !!!!! WARNING !!!!!

  The SHDesign's DLM (which we don't have source for) and the NSCL RTOS
  code both use these structures.  Changing these structures in ANY way
  (including changing the size of items within it) WILL almost certainly
  cause great havoc.
-----------------------------------------------------------------------------*/

#ifndef __DLMIII_H
#define __DLMIII_H

#ifdef WIN32
#pragma pack(1)
extern "C" {
#endif

#include "ZWIDblock.h"
#include "SoftoolsIDBlock.h"

#define PARAM_START 0xF000l


// !!!WARNING!!!! Don't even THINK about changing these!!!!
// Other code (that we don't have the source for) depends on these values !!!
#define  MAX_BLK_NUM   1024
#define  BRD_NAME_LEN   128
#define  MAX_FILES       12
#define  EOF_BLK     0xffff


typedef struct {
  char name[100];
  unsigned short startblock;
  unsigned short flags;
  unsigned long size;
  unsigned long date;
} FILE_ENTRY;

//--- values for FILE_ENTRY.flags ---
#define FILE_ACTIVE  0x0001
#define FILE_SFLASH  0x0002
#define FILE_DCTIME  0x0004


// now used for dir list.
typedef struct {
  char name[16];
  unsigned long size;   // #bytes
  unsigned long date;   //
} SFILE_ENTRY;


typedef struct {
  unsigned char port,       // SxDR
                cs_port,    //
                cs_mask;    // bit to set to turn on
  unsigned long start_block, size;  // use as a circular buffer.
  unsigned short block_size;
} sflash_config;


//===== Basically a File Allocation Table (FAT) plus other config data =====
typedef struct {
  char  board_name[BRD_NAME_LEN];
  unsigned long  IPAddress;  // baud rate on serial version
  unsigned short int  port;  // will be serial port for serial version
  unsigned char  flags, init;
  FILE_ENTRY  files[MAX_FILES];
  unsigned char  primary, backup;  // index(+1) for boot image files
  unsigned char  init_addr[40];
  unsigned char  init_data[40];
  unsigned short  num_blocks, sector_size;
  unsigned short  nextBlock[MAX_BLK_NUM]; // 0 indicates a free block
  char  modem_init[76];
  unsigned long  comm_baud;
  unsigned long  ser_baud;
  char phone_number[32];
  char comm_flags;
  char comm_port;
  char ser_port, ser_flags;
  unsigned long version;
  sflash_config sflash;
  ZWorldIDBlock ZWIDBlock;
  SoftoolsIDBlock STIDBlock;
  unsigned long key;
  unsigned long ZWUserBlockAddr;
} CONFIG_TABLE;

#define CONFIG_SFLASH     1
#define CONFIG_SFLASH2    2
#define CONFIG_SF_FAT     4
#define CONFIG_MODEM      8
#define CONFIG_SOFTOOLS  16 // use ANSI times on files


typedef struct {
  unsigned long  Key;
  unsigned long  IPAddress;
  unsigned short  Port;
  unsigned short  Cmd, Flags;
  char IDString[128-14];
} RUN_INFO;

#define CONFIG_KEY  0xFACE55aal
#define PRIMARY_KEY 0xFACE5533l
#define BACKUP_KEY  0xFACE55ccl
#define INIT_KEY    0xFACEAA55l


#ifdef WIN32
};
#pragma pack()
#endif
#endif
