//----- SHD_DLM.c ----- 12/07/10 -----(xx/xx/xx)-----

//--- original code from SHDesigns ---

//--- modifications and reformatting by Mark Davis, NSCL ---


#include <stcpip.h>

#include <SHD_DLM.h>
#include <DLMiii.h>

#include "string.h"

//#define DLM_SERIAL

//#include "../include/dlmiii.h"
#include "serial.h"

#include <sys/hdwDefs.h>

#include <sys/farprintf.h>


#define  RUN_INFO_ADDR   (0x80000l - 128l)

#define DLM_RW

//-----------------------------------------------------------------------------
//  NSCL change log (most recent first):
//
//
//  2010-12-07 - Mark Davis
//               - Added support for UDPDNLD_MODE_SOFT_RESET bit in mode value
//                 for calls to DLM_Reboot().
//
//  2010-06-29 - Mark Davis
//               - Commented out functions not used (or currently compatible
//                 with) DEBROS.
//
//  2010-03-11 - Mark Davis
//               - Added DLM_RunKey() to determine which boot image is running
//                 and if we restarted due to the RDLM loading us (vs. some
//                 bug that caused us to run the startup code again without
//                 actually rebooting).
//               - DLM_Init() no longer returns the value indicating which
//                 image was loaded.
//
//  2009-12-10 - Mark Davis
//               - DLM_Init() now longer opens a socket.  This is to eliminate
//                 possible problems with the DLM code opening sockets that
//                 are outside of DEBROS control.  Since DLM_tick() is no
//                 longer called for the same reason, this does not present
//                 a problem.
//
//  2009-03-19 - Mark Davis
//               - DLM_Init() now returns a negative value if an error occurs
//                 (rather than a positive #) or one of the values
//                 UDPDNLD_MODE_RUN_PRIMARY or UDPDNLD_MODE_RUN_BACKUP to
//                 indicate which of the boot image files was loaded.
//
//-----------------------------------------------------------------------------

/*-----------------------------------------------------------------------------
SHDesigns Resident DLM library Change log:


Version 1.1:  Date: 05-08-2006
  - Added read/write support to existing files.
  - Seek support added.

Version 1.5B  Date: 7-27-2006 (version # synced with DC lib, there was no 1.2-1.4)
  - Added Support for reading/writing file within a user program
  - Fixes for configuration data
  - Final changes for variable config_table position
  - Added option to reboot application or reboot to DLM.
  - Added normal and verbose libs.
  - fixed ffile_seek issue on files that a re multiple of 4k.

Version 1.6A: Date: 1-6-2007:
  - Added serial support

-----------------------------------------------------------------------------*/

extern CONFIG_TABLE far * config_table;
extern unsigned short sector_size;
int DLM_SerEnable = 1;

static struct {
  unsigned udpdl_port;
  char * id_string;
  udp_Socket udp_dl_sock;
  char flags;
} _dl_data = {DNLD_UDP_PORT, ""};

extern unsigned long param_start;
int init_file_data(void);


//-----------------------------------------------------------------------------
#include <sys/types.h>
#include <unistd.h>

#include <busybox.h>

#include <sys/serial.h>


//-----------------------------------------------------------------------------
/*
void dump_Run_Info(ulong addr, int count)
{
  uchar  i, x;
  int  remain;
  char far *ptr;
  char  msg[80], buf[16];

  printf("------------------------------------------------\r\n");
  ptr = (char far *)addr;
  for (remain=count; remain>0; remain-=16)  {
    x = (uchar)((remain >= 16) ? 16 : remain);
    for (i=0; i<x; ++i)  buf[i] = *ptr++;
    fmtHexLine(msg, addr, buf, x);
    strcat(msg, "\r\n");  printf(msg);  flushConsole();
    addr += 16;  if (remain < 16)  break;
  }
  printf("------------------------------------------------\r\n\n");

  flushConsole();
}
*/

//-----------------------------------------------------------------------------
//  Return the key value from the RDLM RUN_INFO structure and set it to 0.
//
//  Return value:
//     0:  This function was called more than once since the RDLM loaded the
//         code.  This is a handy way to tell if we started up properly, or
//         somehow cstart() and/or main() was executed again without being
//         reloaded by the RDLM.
//
//     UDPDNLD_MODE_RUN_PRIMARY:
//         The primary boot image was loaded
//
//     UDPDNLD_MODE_RUN_BACKUP:
//         The backup boot image was loaded
//-----------------------------------------------------------------------------
int DLM_RunKey(void)
{
  unsigned long key;
  int retval;

//--- get flag indicating which boot image we are running ---
  retval = 0;  key = ((RUN_INFO far *)RUN_INFO_ADDR)->Key;

//  dump_Run_Info(RUN_INFO_ADDR, sizeof(RUN_INFO));  //tdebug

  if (key == PRIMARY_KEY)  retval = UDPDNLD_MODE_RUN_PRIMARY;
  if (key == BACKUP_KEY)   retval = UDPDNLD_MODE_RUN_BACKUP;

  ((RUN_INFO far *)RUN_INFO_ADDR)->Key = 0;

  return retval;
}


//-----------------------------------------------------------------------------
//  Initialize DLM-related functionality and the flash file system
//
//  Return value:
//     < 0 if an error
//     0 otherwise
//-----------------------------------------------------------------------------
int DLM_Init(char *str)
{
  _dl_data.flags = 0;
  _dl_data.id_string = str;

  if (str == NULL)  _dl_data.id_string = "Rabbit Board UDP Download";

/*
  if (_dl_data.flags & DL_SOCKET_OPEN) {
    sock_close(&_dl_data.udp_dl_sock);  _dl_data.flags = 0; }

  if (!udp_extopen(&_dl_data.udp_dl_sock, IF_ANY, _dl_data.udpdl_port,
                   -1L, -1, NULL, 0, 0))  return -1;
  _dl_data.flags |= DL_SOCKET_OPEN; // tell future calls socket is open.
*/

  init_file_data();

#ifdef DLM_SERIAL
  if (!serXopen(config_table->comm_baud, config_table->ser_flags >> 1))
    printf("DLM serial port failed to open");
#endif

  return 0;
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void DLM_SetPort(unsigned port)
{
  _dl_data.udpdl_port = port;
}


//-----------------------------------------------------------------------------
/* START FUNCTION DESCRIPTION ********************************************
int DLM_Reboot(int mode);                       <SHD_DLM.LIB>

PARAM: mode
  UDPDNLD_MODE_RUN_DLM    - Run the resident DLM
  UDPDNLD_MODE_RESTART    - Reboot and run Primary
  UDPDNLD_MODE_RUN_BACKUP - Reboot and run Backup

  UDPDNLD_MODE_SOFT_RESET - OR'd to above, causes reboot w/o CPU reset

  UDPDNLD_MODE_RUN_OK     - Tell DM we started up properly
             (otherwise it will run backup on next reboot)

END DESCRIPTION **********************************************************/
//-----------------------------------------------------------------------------
extern unsigned long _xmemEnd;

int DLM_Reboot(int mode)
{
  unsigned long raddr = RUN_INFO_ADDR;
  int retval = UDPDNLD_MODE_RUN_PRIMARY;
  bool  doSoftReset, justReportingOK;
  RUN_INFO run_info;


  fmemcpy(&run_info, (char far *)raddr, sizeof(RUN_INFO));

//  dump_Run_Info(raddr, sizeof(RUN_INFO));  //tdebug

  if (run_info.Key == BACKUP_KEY)  retval = UDPDNLD_MODE_RUN_BACKUP;

  fmemset(&run_info, 0, sizeof(run_info));

  justReportingOK = (mode == UDPDNLD_MODE_RUN_OK);

  doSoftReset = (!justReportingOK AND (mode & UDPDNLD_MODE_SOFT_RESET));

  mode &= ~(int)UDPDNLD_MODE_SOFT_RESET;  // clear UDPDNLD_MODE_SOFT_RESET bit

  switch (mode)  {
    case UDPDNLD_MODE_RUN_DLM:
      run_info.Key = CONFIG_KEY;
      run_info.Port = _dl_data.udpdl_port;
      strcpy(run_info.IDString, _dl_data.id_string);
      run_info.IPAddress = htonl(gethostid());
      break;

    case UDPDNLD_MODE_RESTART: // 0 will just restart
      run_info.Key = 0; // will cause primary to run
      break;

    case UDPDNLD_MODE_RUN_BACKUP:
      run_info.Key = PRIMARY_KEY; // Primary WAS running, try backup next time
      break;
  }
  fmemcpy((char far *)raddr, &run_info, sizeof(RUN_INFO));

  #ifdef VERBOSE
    printf("Set key to %lX at %lx\n", run_info.Key, raddr);
  #endif

  if (!justReportingOK)
    if (doSoftReset)  softReset();  else  hardReset();

  return retval;
}


//-----------------------------------------------------------------------------
/* START FUNCTION DESCRIPTION ********************************************
int DLM_Tick();                       <SHD_DLM.LIB>

SYNTAX: int DLM_Tick();

PARAMETERS:

KEYWORDS: download

DESCRIPTION: Checks for download request and processes Queries.

RETURN VALUE: 0 = if no download request
              1 = if Download requested
Note: When this function returns 1 the user could could
shut down user functions like interrupts and threads

After this routine returns 1, at some later call, the DLM
will be executed and any user code may be lost. This return allows user
code to exit gracefully. I.e. flush buffers, close connections. Turn off
power to devices.

Note: downloads can not be run under the debugger, the DC IDE will get confused
 and reset the board.

END DESCRIPTION **********************************************************/
//-----------------------------------------------------------------------------
int process_dlm(UDP_DBG_HEADER * up, int * retval);

//static int reboot_dlm;
//static word port;
//static unsigned long ip;

/*
//-----------------------------------------------------------------------------
int DLM_Tick(void)
{
  auto UDP_DBG_HEADER * up;
  auto int len;
  auto int retval = 0;
  auto char buff[1500]; // allow room for max code block

  up = (UDP_DBG_HEADER *)buff;

  if (_dl_data.flags & DL_RUN_PENDING) // no need to clear flags, we're history anyway.
    DLM_Reboot(reboot_dlm ? UDPDNLD_MODE_RESTART : UDPDNLD_MODE_RUN_DLM);

  len = udp_recvfrom(&_dl_data.udp_dl_sock, buff, sizeof(buff), &ip, &port);

  if (len > 0) {
    retval = 0; // tell handler we are UDP
    if (process_dlm(up, &retval)) // send ack or data
      udp_sendto(&_dl_data.udp_dl_sock, buff,
                 UDP_DBG_HEADER_SIZE + up->length, ip, port);
    return retval;
  }

#ifdef DLM_SERIAL
  if (!DLM_SerEnable)  return 0;

  len = ser_get_packet((UDP_SDBG_HEADER *)buff);
  if (len > 0)  {
    retval = 1; // tell handler we are serial
    up = (UDP_DBG_HEADER *)(buff + 1);
    if (process_dlm(up, &retval)) {
      // send ack or data
      send_sack((UDP_SDBG_HEADER *)buff);
    }
    return retval;
  }
#endif

  return retval;
}

//-----------------------------------------------------------------------------
//
// *retval will be on if called from serial
//
//-----------------------------------------------------------------------------
int process_dlm(UDP_DBG_HEADER * up, int * retval)
{
  int  length, sendack = 0;
#ifdef DLM_RW
  int err = 0;
  static char far *sector_buff = (char far *)0l;
  static unsigned long  address, block_addr = -1;
  static unsigned  sector_off, doff, offset, blength;
#endif
  auto char  mac[6];


  if (!sector_buff) sector_buff = xalloc(4096);
  if (!sector_buff) return -1;

  length = up->length;

  switch (up->cmd)  {

#ifdef DLM_RW
    case  UDPDNLD_FLUSH:
        // need toflush partial block?
       if ((block_addr >= 0) && sector_off)  {
       #ifdef VERBOSE
         printf("Write last block. Address = 0x%8lX, len=%d\n",
                block_addr, sector_off);
       #endif
        _FlashUnlock = ~block_addr;

        err -= WriteFlash((FlByte const far *)block_addr, sector_buff,
                          sector_off);
      }
      *retval = 1;  goto block_err;

    case UDPDNLD_DOWNLOAD_FLASH:
      #ifdef VERBOSE
        printf("Download. Address = 0x%8lX, len=%d\n", up->address, up->length);
      #endif
      address = up->address & ~((unsigned long)sector_size - 1);  // make block #
      doff = 0;

      // if diff block and partial block, write current
      if ((block_addr >= 0) && (address != block_addr) && sector_off)  {
        _FlashUnlock = ~block_addr;
        err -= WriteFlash((FlByte const far *)block_addr, sector_buff, sector_size);
      }
      block_addr = address;

      if (up->status !=   // prevent overwrite of code
          UDPDNLD_STATUS_DLM_OVERWRITE && block_addr < param_start) {
        err = 1;  goto block_err;  }

      if (up->address >= param_start && up->address < (param_start + 4096))
       fmemcpy(((char far *)config_table) + (up->address - param_start),
                 up->data, up->length);
      offset = (int)(up->address&(sector_size - 1));    // mask off

      if (offset == 0) {
        ReadFlash((FlByte const far *)address, (FlByte far *)sector_buff, sector_size);
        block_addr = address;
      }

      while (length > 0) {
        blength = length;
        if ((blength + offset) > sector_size)   // sectors<512 bytes;
          blength = sector_size - offset; // offset SHOULD be 0
        fmemcpy(sector_buff + offset, &up->data[doff], blength);   // copy to buff
        doff += blength;  length -= blength;  offset += blength;
        if (offset >= sector_size) {
          _FlashUnlock = ~block_addr;
          err -= WriteFlash((FlByte const far *)block_addr, sector_buff, sector_size);
          if (err == 0) { block_addr += sector_size;  offset = 0; }
        }
      }
      *retval = 1;  sector_off = offset;

  block_err:
      sendack = 1;  up->length = 0;
      up->status = (err==0) ? UDPDNLD_STATUS_ACK : UDPDNLD_STATUS_NACK;    // resend rx packet as ack/nack
      break;
#endif

    case UDPDNLD_READ:
      if (up->length == 0)  return 0; // check valid, no data: spurious packet
      if (!param_start)  init_file_data();
      if (up->address >= param_start && up->address < (param_start + 4096))
        fmemcpy(up->data,
                ((char far *)config_table) + up->address - param_start,
                up->length);
      else
          ReadFlash((char far *)up->address, paddr(up->data), up->length);
      *retval = 0;  sendack = 1;  up->status = UDPDNLD_STATUS_ACK;
      #ifdef VERBOSE
        printf("Read from %lx, %d bytes\n", up->address, up->length);
      #endif
      break;

    case UDPDNLD_QUERY3:
      if (up->length != 0) return 0; // check valid, no data: spurious packet
      length = strlen(_dl_data.id_string);
      if (length > 79) length = 79;
      memcpy(up->data, _dl_data.id_string, length);
      up->data[length] = '\0';  length++;
      pd_getaddress(0, mac);
      sprintf(up->data + length, " %02x:%02x:%02x:%02x:%02x:%02x",
              mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
      length += strlen(up->data + length);
      #ifdef DLM_RW
        up->status = UDPDNLD_STATUS_USER_RW; // tel downloader we can handle read/write
      #else
        up->status = UDPDNLD_STATUS_USER_CODE; // tell downloader to send us loader
      #endif
      if (*retval)
        up->address = 600;    // serial buff is 600
      else
        up->address = ETH_MTU;  // used by downloader
      up->length = length + 1;  sendack = 1;
      #ifdef VERBOSE
        printf("Query from %lx port %d\n", ip, port);
      #endif
      break;

    // this is only called if loadein is not in FLASH
    case UDPDNLD_REBOOT: // off we go!
      reboot_dlm = up->status;
      _dl_data.flags |= DL_RUN_PENDING;  sendack = 1;  *retval = 2;
      break;

    case UDPDNLD_SET_SPECIFIC_IP:
      pd_getaddress(IF_ETH0, mac);
      if (memcmp(&up->data, mac, 6) != 0) return 0;
      // fall into set IP.

    case UDPDNLD_SET_IP:
      sethostid(up->address);
      // ip has changed so we need to close and reopen socket.
      sock_close(&_dl_data.udp_dl_sock);
      up->status = UDPDNLD_STATUS_ACK;
      udp_extopen(&_dl_data.udp_dl_sock, IF_ETH0, _dl_data.udpdl_port,
                  -1L, -1, NULL, 0, 0);
      break;

    case UDPDNLD_SET_RTC:
      writeRTC(up->address);
      up->length = 0;
      up->status = UDPDNLD_STATUS_ACK;  sendack = 1;  *retval = 3;
      break;

    case UDPDNLD_GET_PARAMS:
      if (!param_start)  init_file_data();
      up->length = 0;  up->address = param_start;
      up->status = UDPDNLD_STATUS_ACK;  sendack = 1;
      break;
  }
  return sendack;
}

//-----------------------------------------------------------------------------
int DLM_SetFallbackIP(unsigned long ipaddr)
{
  config_table->IPAddress = ipaddr;
  _FlashUnlock = ~param_start;
  WriteFlash((FlByte const far *)param_start, (FlByte const far *)config_table,
             sector_size);
  return 0;
}
*/

//-----------------------------------------------------------------------------

