//----- SHD_DLM.h ----- 01/26/11 -----(xx/xx/xx)-----

//----- original code from SHDesigns -----

//--- additions, modifications, and reformatting by Mark Davis, NSCL ---

//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//
//  2011-01-26 - Mark Davis
//               - getIPAddr() and setIPAddr() changed to readIPAddr() and
//                 writeIPAddr()
//
//-----------------------------------------------------------------------------

#ifndef __SHD_DLM
#define __SHD_DLM


#include <sys/socket.h>

#include <DLMIII.h>



#define  DOPEN_FILES   4  // max # of flash files open at once


//===== Info about a currently opened file =====

typedef struct {
  unsigned fatIdx;         // Index in to File Table
  unsigned curBlkNum;      // sector # corresponding to data in R/W buf (0=none)
  unsigned long pos;       // offset from start for next read/write operation
  unsigned long sectorPos; // offset from start of first byte in sector_buff
  char far *sector_buff;   // R/W buffer for current sector
  char far *buff_ptr;      // points to addr within sector_buff
  int flags;               // see below
} FILE_DATA;

#define FFLAG_OPEN    0x01  // FILE_DATA struct is in use by an open file
#define FFLAG_WRITE   0x02  // R/W buffer contains newer data than the flash
#define FFLAG_UPDATED 0x04  // The FAT needs to be flushed to flash


//--- list of open references to files ---
//--- The ffile "handles" are indexes in to this list ---
extern FILE_DATA  openFiles[DOPEN_FILES];



//===== definitions related to the Download Manager interface =====

#define DLM_VERSION "1.6A"

#define DNLD_UDP_PORT 2000  // arbitrary

typedef struct {
  unsigned char cmd, status;
  unsigned short length;  // data size
  unsigned long address;  // write address
  unsigned char data[ETH_MTU-8];  // can be up to 512 bytes.
} UDP_DBG_HEADER;

#define UDP_DBG_HEADER_SIZE (sizeof(UDP_DBG_HEADER)-ETH_MTU+8)

// serial support
#if defined(SERIAL_A) || defined(SERIAL_B) || defined(SERIAL_C) || defined(SERIAL_D) || defined(SERIAL_E) || defined(SERIAL_F)
#define DLM_SERIAL
#endif

typedef struct {
  unsigned char flag, cmd, status;
  unsigned short length;  // data size
  unsigned long address;  // write address
  unsigned char data[ETH_MTU-8];  // can be up to 512 bytes.
} UDP_SDBG_HEADER;

#define UDP_SDBG_HEADER_SIZE (sizeof(UDP_SDBG_HEADER)-ETH_MTU+8)

#define FRAG_MASK 0xff20

// command definitions
#define UDPDNLD_QUERY           1 // query for all boards (broadcast) (old DLM)
#define UDPDNLD_DOWNLOAD_RAM    2 // download RAM code
#define UDPDNLD_XMEM_SIZE       3 // xmem size needed for RAM code
#define UDPDNLD_DOWNLOAD_FLASH  4 // download FLASH code (data len=0 when done)
#define UDPDNLD_RUN             5 // run download RAM code
#define UDPDNLD_REBOOT          6 // restart with new code
#define UDPDNLD_SET_IP          7 // set the board to diff IP
#define UDPDNLD_SET_SPECIFIC_IP 8 // set the board to diff IP
#define UDPDNLD_GET_USERBLOCK   9
#define UDPDNLD_READ           10 // read memory
#define UDPDNLD_FLUSH          11 // flush last block
#define UDPDNLD_QUERY3         12 // query for all boards (broadcast)
#define UDPDNLD_SET_RTC        15
#define UDPDNLD_GET_PARAMS     16
#define UDPDNLD_NULL          127 // do nothing command to resolve ARP

// query status
#define UDPDNLD_STATUS_NULL      0 // not set
#define UDPDNLD_STATUS_NEED_CODE 1 // board needs code downloaded to RAM
#define UDPDNLD_STATUS_HAS_CODE  2 // board has code
#define UDPDNLD_STATUS_RAM_CODE  3 // ram code already running
#define UDPDNLD_STATUS_NOMEM     4 // no xmem available
#define UDPDNLD_STATUS_REBOOT    5 // rebooting

// download status
#define UDPDNLD_STATUS_ACK               6 // download acknowledge
#define UDPDNLD_STATUS_NACK              7 // download fail acknowledge
#define UDPDNLD_STATUS_SEQUENCE          8 // dld sequence error
#define UDPDNLD_STATUS_RAM_CODE_IN_XMEM  9 // RAM loader in flash
#define UDPDNLD_STATUS_FRAG             10 // fragmented packet received

// new code status
#define UDPDNLD_STATUS_USER_CODE        11 // running user code
#define UDPDNLD_STATUS_USER_RW          12 // User prog with R/W support
#define UDPDNLD_STATUS_DLM_OVERWRITE    13

#define UDPDNLD_MODE_RUN_DLM        0x00
#define UDPDNLD_MODE_RESTART        0x01
#define UDPDNLD_MODE_RUN_BACKUP     0x02
#define UDPDNLD_MODE_RUN_PRIMARY    0x03
#define UDPDNLD_MODE_RUN_OK           -1

// for use with mode value for DLM_Reboot()
#define UDPDNLD_MODE_SOFT_RESET     0x10


int  DLM_Init(char * str);
int  DLM_Tick(void);
void DLM_SetPort(unsigned port);
int  DLM_Reboot(int mode);

#ifdef DLM_SERIAL
  extern int DLM_SerEnable;
  extern FILE * DLM_serio;
#endif


//
// file I/O functions
//

#define FERR_NOTFOUND      -1
#define FERR_TOOMANYFILES  -2
#define FERR_NOXMEM        -3
#define FERR_FILE_EXISTS   -4
#define FERR_SEEK          -5



// open existing file for reading/writing
int ffile_open(char * name);

// read from file, returns # bytes read or < 0 if error
int ffile_read(int handle, char far *buff, int bytes);

// write to file, returns bytes written or < 0 if error
int ffile_write(int handle, char far *buff, int bytes);

// create a new file for writing
int ffile_create(char * name);

// delete a file
int ffile_delete(char * name);
int ffile_delete_index(int index);

// return current file position
long ffile_tell(int handle);

// seek within file
// returns 0 if no error, FERR_SEEK if beyond file
int ffile_seek(int handle, long pos, int mode);

#define FSEEK_SET     0
#define FSEEK_CUR     1
#define FSEEK_END     2


// file size of open file
long ffile_size(int handle);

// file size by name
long ffile_sizen(char * name);

// close a file opend for read or write
void ffile_close(int handle);

// get entry number for a file
int ffile_index(char * name);

// get index # on open file (use on the set primary and backup calls)
int ffile_handle2Index(int handle);


// set entry number as primary or backup or init
int ffile_setPrimary(int index);
int ffile_setBackup(int index);
int ffile_setInit(int index);

// return free space in flash
long int  ffile_free(void);

// create a link to a FAT file, returns index
int ffile_FATLink(char * fname);

// get file name
void ffile_name(unsigned int index, char * name);
void ffile_rename(unsigned int index, char * name);

int ffile_getPrimary(void);
int ffile_getBackup(void);
int ffile_getInit(void);

int DLM_SetFallbackIP(unsigned long ipaddr);

// dl_flags bits
#define DL_RUN_PENDING 1
#define DL_SOCKET_OPEN 4



//========== NSCL additions ==========

#include <sys/types.h>


int  DLM_RunKey(void);  // 1st call tells if PRIMARY or BACKUP image is running


FILE_ENTRY far *ffile_info(unsigned idx);
int  ffile_fstat(int handle, struct stat *fsbuf);

long ffile_unread(int handle);

unsigned long  readIPAddr(void);
int  writeIPAddr(unsigned long ipAddr);

char *ffile_getPrimaryName(char *name, int nameLen);
char *ffile_getBackupName(char *name, int nameLen);

int ffile_rename_(char const *oldpath, const char *newpath);

int ffile_makePrimary(const char *fileName);
int ffile_makeBackup(const char *fileName);

int  ffile_utime(const char *fileName, const struct utimbuf *buf);

int   ffile_setName(char *newName);
char *ffile_getName(char *name, int maxLen);

int ffile_fsck(bool repair);



#endif
