//----- timerB.h ----- 06/11/12 -----(xx/xx/06)-----

#ifndef _timerBH_
#define _timerBH_


//--- number of bytes pushed on to the stack by _TimerBISR() when it calls ---
//--- timer_interrupt() (includes the return address with XPC value) ---

#define  TimerBISR_PushCount  23



//-----------------------------------------------------------------------------
//  TimerB interrupt handler.  Original version from SHDesigns TimerB sample
//  code (http://www.shdesigns.org)
//-----------------------------------------------------------------------------
void near _nostep _TimerBISR(void);


//-----------------------------------------------------------------------------
// SYNTAX:   int select_timer(int usec, char intLevel)
//
// DESCRIPTION:
//   Call external func  sched_tick() to be called every SchedMS milliseconds
//
//   Notes:
//    1. Periodic call to user routine every timer int
//    2. Supports times from about 20us to 1500 seconds with 11.0592mHz xtal
//       (22.1 clock)
//    3. Interrupts faster than 20,000/sec start to become too much overhead
//       for use
//    4. select_timer() must be called to initialize the ISR
//    5. Uses bios freq_divider for calcs
//    6. If <100us, best to modify library asm isr. (inline code, rather than
//       a call to the external func)
//    7. Note serial port baud rates <2700 effect timer, set up serial ports
//       first
//
//    Calls external function:   void _nostep sched_tick(void);
//
// RETURN VALUE: non-0 if out of range
//-----------------------------------------------------------------------------
int select_timer(unsigned long intsPerSec, char intlevel);


//-----------------------------------------------------------------------------
//  SYNTAX:   float TimerBRate()
//
//  DESCRIPTION:   Return actual ISR time in microseconds
//
//  RETURN VALUE: float, microseconds per interrupt
//-----------------------------------------------------------------------------
float TimerBRate(void);


//-----------------------------------------------------------------------------
//  SYNTAX:   void TimerBUninit()
//
//  DESCRIPTION:  Stop the timer
//
//  RETURN VALUE: None.
//
//-----------------------------------------------------------------------------
 // # times the ISR will be invoked before the external func is called.
extern unsigned int _timerb_scale;

void TimerBUninit(void);


//-----------------------------------------------------------------------------
//  The external function called at specified intervals
//-----------------------------------------------------------------------------
void sched_tick(void);


#endif
