//--- bytebuf.h --- 05/26/05 ---(05/06/05)---

#ifndef _bytebufH_
#define _bytebufH_


//-----------------------------------------------------------------------------
//
//  Copyright (c) 2003-2006 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//-----------------------------------------------------------------------------
//  Data type for managing a circular buffer in extended memory
//-----------------------------------------------------------------------------
//
//  2006-05-26 - Mark Davis
//               - ported from ZWorld compiler to Softools compiler
//
//  2005-06-07 - Mark Davis -
//               - Added Buf_Lock() and Buf_Unlock() functions
//
//  2005-05-06 - Mark Davis -
//               - Moved existing ByteBuf pieces in to separate module
//
//-----------------------------------------------------------------------------


typedef struct  {
  ulong  bufAddr;
  int    bufSize;
  int    inIdx, outIdx;
  ulong  lockCount;
  uchar  initFlag;   // set to 0xA5 if buf has been initialized
}  ByteBuf;


void  Buf_Init(ByteBuf *buf, int bufSize);


void  Buf_Flush(ByteBuf *buf);

#define  Buf_Size(bb)  ((bb)->bufSize)

int Buf_Used(ByteBuf *toBuf);

int Buf_Avail(ByteBuf *toBuf);

int Buf_Add(ByteBuf *toBuf, char *fromBuf, int fromCount);

int Buf_Copy(ByteBuf *fromBuf, int offset, char *toBuf, int copyCount);

int Buf_LineLen(ByteBuf *fromBuf, int offset);

int Buf_Remove(ByteBuf *fromBuf, char *toBuf, int removeCount);

ulong Buf_Lock(ByteBuf *buf);
ulong Buf_Unlock(ByteBuf *buf);
#define  Buf_Locked(buf)  ((buf)->lockCount)

int  Buf_Add_Short(ByteBuf *toBuf, int val);
int  Buf_Add_Long(ByteBuf *toBuf, long val);

int Buf_Remove_Short(ByteBuf *byteBuf);
long Buf_Remove_Long(ByteBuf *byteBuf);



#endif
