//----- sys/modbus.h ----- 12/05/12 -----(05/08/06)-----

#ifndef _sys_modbusH_
#define _sys_mdobusH_

//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-12-05 - Mark Davis
//               - Moved MB_cmdBitsStd() and MB_timeZone() from modbus.c
//                 to sharedApplic.c
//
//  2012-07-18 - Mark Davis
//               - Changes to PV_Desc and related macros to support locking
//                 and PVs without a separate read buffer.
//
//  2012-04-xx - Mark Davis
//               - Changed modbusScanTables() to pvInitLists()
//               - Got rid of pvToRegInfo()
//               - Added CopyFunc and pvCopyVal()
//               - Added PV_DESC() macro
//
//  2012-03-27 - Mark Davis
//               - Began changes to support PV types defined by structures
//                 other then RegInfo.
//
//  2012-02-23 - Mark Davis
//               - Added numPVs and pvList.  Changed numC0Coils, numR3Regs,
//                 and numR4Regs to type uint.
//
//  2012-02-14 - Mark Davis
//               - Changed references to pid in the PV functions to just id
//                 to avoid confusion with process IDs.  For the same reasons
//                 I made the following func name changes:
//                    pvGetPID() is now pvGetID()
//                    pidToRegInfo() is now pvToRegInfo()
//
//  2012-02-01 - Mark Davis
//               - Added comments and DBE_MASK_BITS
//
//  2011-04-22 - Mark Davis
//               - Added restoreIfRegValue() for other modules to call if they
//                 have a value to write to memory that may be a register.
//
//  2011-04-19 - Mark Davis
//               - Added MB_setACoil() and MB_resetACoil()
//      ...
//
//  2005-08-06 - Mark Davis
//               - original version
//
//-----------------------------------------------------------------------------

//#include <sys/types.h>


#define  MAX_PV_NAME_LEN    64



//--- values for RegInfo.wrMask ---
#define  REG_RO   0  // no writes allowed
#define  REG_RW   1  // writes allowed by all clients
#define  REG_WL   2  // writes allowed by local clients only

//--- values for RegInfo.dataType ---
#define  DT_S16        0     // 16-bit signed integer
#define  DT_U16        1     // 16-bit unsigned integer
#define  DT_S32        2     // 32-bit signed integer
#define  DT_U32        3     // 32-bit unsigned integer
#define  DT_F32        4     // 32-bit IEEE Float
#define  DT_STR        5     // null-terminated string (for conversions only)
#define  DT_U8         6     // used outside of ModBus interface
#define  DT_S8         7     // used outside of ModBus interface
#define  NUM_MB_TYPES  8

extern const char *typeDesc[NUM_MB_TYPES];

//--- typedefs for each dataType ---
typedef  unsigned char   U8;
typedef  char            S8;
typedef  short           S16;
typedef  unsigned short  U16;
typedef  long            S32;
typedef  ulong           U32;
typedef  float           F32;

//--- # of registers for each dataType ---
#define  NR_S16   1
#define  NR_U16   1
#define  NR_S32   2
#define  NR_U32   2
#define  NR_F32   2


#define C0_ValDef(name)    name[NumC0Chans_]
#define C0xValDef(name)    name

#define R3_ValDef(name)    name[NumR3Chans_]
#define R3xValDef(name)    name

#define R4_ValDef(name)    name[NumR4Chans_]
#define R4xValDef(name)    name


#define C0_ValPtr(name)    ((uchar *)(name))
#define C0xValPtr(name)    ((uchar *)(&(name)))

#define R3_ValPtr(name)    ((int *)(name))
#define R3xValPtr(name)    ((int *)(&(name)))

#define R4_ValPtr(name)    ((int *)(name))
#define R4xValPtr(name)    ((int *)(&(name)))



//--- One RegInfo or CoilInfo structure per logical value ---

#define  SetCoilFuncParams  \
   void *info, uchar chanNum, uchar newVal, fd_t eventsfd
typedef bool (*SetCoilFunc)(SetCoilFuncParams);

typedef struct  {
  uchar       *coilVal;     // ptr to true/false value
  uchar        coilNum;     // # of coil
  const char far *name;     // ptr to string with coil name
  SetCoilFunc  setCoilFunc; // to validate value, generate event, etc
  byte         flags;       // (see below)
} CoilInfo;


#define  _SetRegFuncParams_  info, chanNum, newVal, eventsfd
#define  SetRegFuncParams  \
                  void *info, uchar chanNum, void *newVal, fd_t eventsfd
typedef bool (*SetRegFunc)(SetRegFuncParams);

//#define  U16VAL  ( *((U16 *)newVal) )
//#define  S16VAL  ( *((S16 *)newVal) )
//#define  U32VAL  ( *((U32 *)newVal) )
//#define  S32VAL  ( *((S32 *)newVal) )
//#define  F32VAL  ( *((F32 *)newVal) )

#define  MB_VAL(type)  ( *((type *)newVal) )


typedef struct  {
  S16         *regVal;      // ptr to first reg in possibly multi-reg value
  U16          regNum;      // # of first reg in possibly multi-reg value
  const char far  *name;    // ptr to string with reg name
  SetRegFunc   setRegFunc;  // to validate value, generate event, etc
  U8           dataType;    // data type (DT_U16, _S16, _U32, _S32, _F32)
  U8           numRegs;     // # of registers that make up the value
  U8           wrMask;      // which processes can change this value?
  byte         flags;       // (see below)
  byte         checkEvents; // CA events to be checked
  byte         eventsOccured; // CA events that occured
  void        *lastChgVal;  // value when last value chg event was recognized
  void        *minChg;      // min chg from lastChgVal for new value chg event
} RegInfo;


#define REG_ARRAY(typ, firstReg, idx)  \
         ( *( (typ *) ((&firstReg##_RegInfo + (idx))->regVal) ) )


//--- values for CoilInfo.flags and RegInfo.flags ---
#define  OnePerChan   0x01  // single/shared instance or 1 per logical device?
#define  LogChanges   0x02  // log changes made via Modbus interface

//--- values for RegInfo.checkEvents and eventsOccured ---
//=== NOTE:  These should match the EPICS CA definitions for efficiency ===
#define  CUR_VALUE    0x00  // used only by pvGetVal() function
#define  DBE_VALUE    0x01  // value changed by at least minChg value
#define  DBE_LOG      0x02  // not supported at present
#define  DBE_ALARM    0x04  // not supported at present
#define  DBE_MASK_BITS  0x07
#define  DBE_PENDING  0x80  // used by DEBROS CA service to indicate the
                            // value still needs to be sent

//--- values for clientType (used in some func calls to handle ---
//--- differences in byte order expected by different client types) ---
#define  CT_MBSerial     0    // ModBus Serial
#define  CT_MBNSCL       1    // ModBus extenstions to NSCL Serial
#define  CT_MBTCP        2    // Open ModBus/TCP
#define  CT_Native       3    // Native code (internal)


//--- One pointer to one of the above structures for each register ---
//--- (i.e. the entry for each register that is part of a multi-register ---
//---  value will point to the same structure) ---

extern CoilInfo *c0Coils[];
extern RegInfo  *r3Regs[];
extern RegInfo  *r4Regs[];



extern uint  C0ChanMult, NumC0Chans,
             R3ChanMult, NumR3Chans,
             R4ChanMult, NumR4Chans;


#define COIL  ((CoilInfo *)info)
#define REG   ((RegInfo *)info)


bool setRegS16(RegInfo *reg, uchar chanNum, S16 newVal);
bool setRegU16(RegInfo *reg, uchar chanNum, U16 newVal);
bool setRegS32(RegInfo *reg, uchar chanNum, S32 newVal);
bool setRegU32(RegInfo *reg, uchar chanNum, U32 newVal);
bool setRegF32(RegInfo *reg, uchar chanNum, F32 newVal);

#define SetMBReg(reg, chNum, typ, val)  setReg##typ(&reg##_RegInfo, chNum, val)

//--- For other modules to call when they need to write to memory that ---
//--- MAY be a register ---
int restoreIfRegValue(void far *curVal, char *newVal, uint count);


extern uint  numC0Coils, numR3Regs, numR4Regs, numNonRegs, numPVs;


#define  MaxWatchSize  16

//--- defs for "SoftScope" values ---
extern uint  numScopeRegs, captureIdx,
             scopeBufSampleSize, bytesPerScopeSample;
extern ulong  captureCount;
extern RegInfo  *scopeList[MaxWatchSize];
extern char far  *scopeBuf, *capturePtr;





bool _nearcall MB_setACoil(SetRegFuncParams);
bool _nearcall MB_resetACoil(SetRegFuncParams);

bool _nearcall roReg(SetRegFuncParams);

bool _nearcall  setS16(SetRegFuncParams);
bool _nearcall  setU16(SetRegFuncParams);
bool _nearcall  setS32(SetRegFuncParams);
bool _nearcall  setU32(SetRegFuncParams);
bool _nearcall  setF32(SetRegFuncParams);
bool _nearcall  setF32_ABS(SetRegFuncParams);


void *parseValue(RegInfo *reg, char *valStr, char *buf);

char *fmtVal(char *buf, uchar type, int minWidth, void *val);

char *fmtRegVal(RegInfo *reg, S16 *useVal, uint chanNum, char *buf);

int pvInitLists(void);

void _task modbusTCPd(int argc, char **argv);



//--- bit masks for cmdBitsStd ---
#define  CMD_REBOOT          0x00000001  // reboot OS
#define  CMD_SHUTDOWN        0x00000002  // shutdown OS and run DLM/bootloader
#define  CMD_SAVE_DEVDATA    0x00000004  // update devData file from RAM copy
#define  CMD_RELOAD_DEVDATA  0x00000008  // reload devData values from file
#define  CMD_SAVE_HDWCFG     0x00000010  // update hdwcfg file from RAM copy
#define  CMD_RELOAD_HDWCFG   0x00000020  // reload hdwcfg values from file
#define  CMD_UPDATE_BOOTIMG  0x00000040  // update boot image



//===== functions and structures to support EPICS CA/PV style interface =====

typedef struct {
  void  *pvDef;    // pointer to struct defining the PV
  uchar  pvType;   // type of struct that pvDef refers to
}  PV_Ref;

//--- values for pvType ---
#define  PVT_REG   1   // pvDef points to a RegInfo structure
#define  PVT_EXT   2   // pvDef points to a PV_Desc structure


extern PV_Ref far *pvList;



typedef struct {
  const char far *name;  // name of the PV (not including the host name)
  uchar  dataType;       // data type (DT_U16, DT_S16, ...)
  uint   size;           // total size in bytes
  uint   elemSize;       // size of each element
  uint   numElems;       // # of elements
  uint   flags;          // see below
} PV_Info;

//--- values for PV_Info.flags ---
#define  PV_RD         0x0001   // read access
#define  PV_WR         0x0002   // write access
#define  PV_SEND_RAW   0x0004   // on-the-fly format conversion impractical

#define  PV_RW    (PV_RD | PV_WR)


typedef int (*PV_CopyFunc)(void *conInfo, void far *from, uint count);


int  pvGetID(char *pvName);

int  pvGetInfo(int id, PV_Info *pvInfo);


int  pvLockVal(int id, int lockMode, int lockType);

//--- values for lockMode ---
#define  LM_Unlock    0   // Release last successful lock
#define  LM_WaitLock  1   // Return only if lock acquired or an error
#define  LM_TryLock   2   // Returns immediately w/ or w/o acquiring lock

//--- values for lockType ---
#define  LT_READ      0   // Read-only lock for calls to (*sendVal)()
#define  LT_WRITE     1   // Write lock for calls to (*rcvVal)()




//--- see values for RegInfo.checkEvents and eventsOccured above for ---
//--- valid values for the event argument ---
int  pvGetVal(int id, uint toType, uint toSize, void *toVal, uint event);

int  pvSetVal(int id, uint fromType, uint fromSize, void *fromVal);

int  pvCopyVal(int id, uint numElems, void *conInfo, PV_CopyFunc copyFunc);


void pvCheckAllEvents(time_t *dateTime, uint *ms, bool anyChg);

int  pvCheckEvents(int id, uint checkEvents);


//--------------- Info for PVs other than ModBus registers ---------------


typedef struct  PV_Desc_  PV_Desc;



#define  PV_OP_Params_  pv
#define  PV_OP_Params   PV_Desc *pv

typedef int (*PV_OP_Func)(PV_OP_Params);



// see pvLockVal() above for lockMode and lockType values
#define  PV_Lock_Params   PV_Desc *pv, int lockMode, int lockType

typedef int (*PVLockFunc)(PV_Lock_Params);



#define  PV_RW_Params  \
              int id, uint numElems, void *conInfo, PV_CopyFunc copyFunc

typedef int (*PV_RW_Func)(PV_RW_Params);




struct PV_Desc_ {
  PV_Info      pvInfo;

  void far    *val;            // loc of copy of value to send to clients
  uint         valSize;        // size of val buf (pvInfo contains cur size)

  uchar        newEvents;      // events since pvCheckAllEvents() last called
  uchar        eventsOccured;  // events to generate subsriptions for

  PV_OP_Func   updateVal;      // func to update newEvents and *val

  PVLockFunc   lockVal;        // func to lock value b4 sendVal/rcvVal calls
  PV_RW_Func   sendVal;        // func to write PV value to client connection
  PV_RW_Func   rcvVal;         // func to read PV value from client connection
};

extern PV_Desc  nonRegPVs[];


#define  PV_DESC(name)  (nonRegPVs + name##_IDX)


#endif

