//----- unistd.h ----- 11/22/13 -----(03/11/06)-----

#ifndef _unistdH_
#define _unistdH_


//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2013-11-22 - Mark Davis
//               - Added restoredLog to externs so applic code could access it.
//
//  2013-08-09 - Mark Davis
//               - Added simulateMemCorruption flag
//
//  2012-06-07 - Mark Davis
//               - Added schedCallsPrevSec (count of # of times schedule() was
//                 called in the last second).
//
//  2012-03-07 - Mark Davis
//               - Made prevLogInfo far (to free up some more root memory)
//
//  2012-03-01 - Mark Davis
//               - Moved IntShim_PushCount from unistd.c to here for use by
//                 other modules (such as the ps command).
//
//  2012-01-03 - Mark Davis
//               - Added getPC() (returns Program Counter for cur task)
//
//  2011-09-09 - Mark Davis
//               - Added LLRF2 device and hardware types
//
//  2011-02-16 - Mark Davis
//               - Added DT_RFAC (device type for RF Ampl Circulator)
//
//  2011-01-05 - Mark Davis
//               - Added REBOOT() to the header file
//
//  2010-12-14 - Mark Davis
//               - Changed secTimerOffset to hdwResetTime
//
//  2010-12-06 - Mark Davis
//               - Added useSoftReset flag and softReset() func
//
//  2010-09-16 - Mark Davis
//               - Added DT_Test for "application" module that contains cmds
//                 to test various OS functionality.
//
//  2010-07-19 - Mark Davis
//               - Added showMemProtection()
//
//  2010-06-02 - Mark Davis
//               - Added computeMemCRCs() and showISRvectors()
//
//  2010-05-13 - Mark Davis
//               - Replaced setBlockTimeout() with msToTicks()
//
//  2010-03-23 - Mark Davis
//               - Expanded set of DeviceID's that force a fixed IP address
//                 from 0x00F0-0x00FF to 0x00E0-0x00FF.
//
//  2010-03-22 - Mark Davis
//               - postMortem() now takes a string parameter so a description
//                 of the cause can be saved in the crash-log file.
//
//  2010-03-05 - Mark Davis (*)
//               - Added WaitOnConsoleBits to list of standard diag msg values.
//                 This allows me to re-enable the older behavior of waiting
//                 for space in serial port A's output buffer when writing to
//                 the system log and the console (port A) is not being used
//                 by a user shell.  The old behavior could easily cause
//                 significant delays for any process writing to the system log.
//
//  2010-02-16 - Mark Davis
//               - Added global variables with code seg start and size values
//
//  2010-02-14 - Mark Davis
//               - The BITSON() and BITSOFF() macros (used to toggle bits set
//                 by the "diag events" command) now sets/clears bits in
//                 forcedDigOutMask and forcedDigOutBits rather than simply
//                 using digitalOut.  This insures that bits set for diagnostic
//                 purposes are not cleared by normal application usage.
//
//  2010-02-13 - Mark Davis
//               - Added push/pop of ip (the Interrupt Priority register) to
//                 INT_SAVE_ALL() and INT_RESTORE_ALL()
//
//  2009-12-18 - Mark Davis
//               - Removed unused wdtTicks global variable (the one actually
//                 used is in the DiagLogInfo struct in fcntl.h)
//
//  2009-08-24 - Mark Davis
//               - changed applicTicks to an unsigned int
//
//  2009-06-23 - Mark Davis
//               - Undid the changes to _EnterCritical_() and _ExitCritical_()
//                 because the compiler apparently does NOT use the copy of SP
//                 saved in IX at the start of a func for all references to
//                 local variables.  The result is that references to local
//                 variables after pushing something on to the stack (that the
//                 compiler does not know about) causes the wrong addresses
//                 to be used.
//
//  2009-06-20 - Mark Davis
//               - _EnterCritical_() and _ExitCritical_() now push and pop the
//                 IP (Interrupt Priority) value on to/off off the stack to
//                 avoid the possibility that the IP register might "overflow".
//
//  2007-06-17 - Mark Davis
//               - began migration from simple periodic polling loops for
//                 blocking to Linux-style wait queues (to reduce wasted CPU
//                 cycles, decrease average latency, improve response times).
//
//  2007-02-20 - Mark Davis
//               - work started on Linux-style method for managing task
//                 switches.  Done to eliminate need for Softools' TurboTask
//                 kernel and allow stacks to be located anywere in memory.
//                 Based on "Simple Task Switcher" code from SH Design's.
//
//-----------------------------------------------------------------------------

#include <linux/sched.h>

#include <Rabbit.h>
#include <Stdlib.h>
#include <dcdefs.h>

#include <linux/limits.h>

#include <sys/types.h>
#include <sys/hdwdefs.h>

#include <fcntl.h>


// # of bytes pushed on the stack (incl the return addr pushed by call to ISR)
#define IntShim_PushCount  6


void listFileSystems(void);  //tdebug

//-----------------------------------------------------------------------------

//--- !!! For diagnostic/performance testing ONLY!!!!!
#define BITSON(bits)  { \
  _EnterCritical_();  \
  forcedDigOutMask |= (bits);  forcedDigOutBits |= (bits); \
  WriteDigOutputs(digitalOut);  \
  _ExitCritical_();  \
}
#define BITSOFF(bits)  { \
  _EnterCritical_();  \
  forcedDigOutMask &= ~(bits);  forcedDigOutBits &= ~(bits);  \
  WriteDigOutputs(digitalOut);  \
  _ExitCritical_();  \
}

#define BITSON_(bits)  { \
  digitalOut |= (bits);  WriteDigOutputs(digitalOut);  \
}
#define BITSOFF_(bits)  { \
  digitalOut &= ~(bits);  WriteDigOutputs(digitalOut);  \
}


//-----------------------------------------------------------------------------
//  Bit masks for real-time tracking of specific events
//
//  Setting bits in these variables causes the corresponding digital outputs
//  to be turned on and off at the beginning and end of the corresponding
//  event.
//-----------------------------------------------------------------------------
extern ulong  TrackApplLoop, TrackScheduler, TrackKernelLock, TrackTimerISR;

#ifdef  EnableTrackBits
  #define  StartInterval(bits)  { if (bits)  BITSON(bits); }
  #define  EndInterval(bits)   { if (bits)  BITSOFF(bits); }
#else
  #define  StartInterval(bits)
  #define  EndInterval(bits)
#endif

//-----------------------------------------------------------------------------

//--- includes kernel, idle, and applic tasks ---
#define  KeyTask(task)  (Task_Idx(task) < 3)


void klockErr(const char far *context);

//--------------------------------------
// Assert/release "the big kernel lock"
//--------------------------------------
void near KLOCK(void);
void near KUNLOCK(void);

//-------------------------------------
//  Is the "big kernel lock" asserted?
//-------------------------------------
#define KLOCKED()  ( _disabled )


#define  _EnterCritical_()  _ipset2
#define  _ExitCritical_()   _ipres

// cannot do the following as this changes the stack pointer which, apparently,
// the compiler does not expect.  Apparently not all local variable references
// use IX (which the compiler sets == to SP at the beginning of the function?)
// I Think these would be OK for blocks of code that do NOT access local
// variables.
//#define  _EnterCritical_()  _asm("push ip", "ipset 3")
//#define  _ExitCritical_()   _asm("pop ip")

/*
#define INT_SAVE()  \
  _asm ( "push ix", "push iy", "push hl", "push de", "push bc", "push af" );
#define INT_RESTORE() \
  _asm ( "pop af", "pop bc", "pop de", "pop hl", "pop iy", "pop ix" );
*/

#define INT_SAVE_ALL()  \
   _asm ( "push ix",   "push iy",   "push hl",   "push de",  "push bc", \
          "push af",   "exx",       "push hl",   "push de",  "push bc", \
          "exx",       "ex af,af'", "push af",   "ex af,af'" );
#define INT_RESTORE_ALL()  \
   _asm ( "ex af,af'", "pop af",    "ex af,af'", "exx",      "pop bc", \
          "pop de",    "pop hl",    "exx",       "pop af",   "pop bc", \
          "pop de",    "pop hl",    "pop iy",    "pop ix" );


void resetWDT(void);

// these can save on extra code and calls to shift functions
#define  MSWL(longVal)  (*(((uint *)&(longVal))+1)) // Most Signif Word of Long
#define  LSWL(longVal)  (*((uint *)&(longVal)))     // Least Signif Word of Long
//#define  LSWL(longVal)  ((uint)(longVal))           // Least Signif Word of Long

#define  MSBI(intVal)   (*(((uchar *)&(intVal))+1)) // Most Signif Byte of Int
#define  LSBI(intVal)   (*(uchar *)&(intVal))       // Least Signif Byte of Int
//#define  LSBI(intVal)   ((uchar)(intVal))           // Least Signif Byte of Int

#define  LSBL(intVal)   (*(uchar)&(intVal))         // Least Signif Byte of Long


#define  AddrOf(func)  _asm( [_HL], "ld hl, " _str_(func) "##", _HL )


#define  _task  _nearcall


#define  EventsFIFO  "fifoEvents"


//-----------------------------------------------------------------------------
//--- for kernel-space use only ---
//-----------------------------------------------------------------------------

#define  DiagFlags     (current->diagFlags)
#define  DiagFlags_    LSWL(current->diagFlags)
#define  DiagFlags__   LSBL(current->diagFlags)

//#define  DiagFlags_   ((uint)current->diagFlags)
//#define  DiagFlags__  ((uchar)current->diagFlags)


//--- the 8 least significant bits are for appl/cmd-specific events ---
#define  ErrorBits          0x0100
#define  SocketEventBits    0x0200
#define  FileEventBits      0x0400
#define  DebugBits          0x0800
#define  EventMsgBits       0x1000
#define  RWCallBits         0x2000
#define  WaitOnConsoleBits  0x4000  // diagOut_() waits on port A if needed
#define  WaitQueueNodeBits  0x8000


#define  ShowErrors        (DiagFlags_ & ErrorBits)
#define  ShowSocketEvents  (DiagFlags_ & SocketEventBits)
#define  ShowFileEvents    (DiagFlags_ & FileEventBits)
#define  ShowDebugMsgs     (DiagFlags_ & DebugBits)
#define  ShowEventMsgs     (DiagFlags_ & EventMsgBits)
#define  ShowRWCalls       (DiagFlags_ & RWCallBits)
#define  WaitOnConsole     (DiagFlags_ & WaitOnConsoleBits)
#define  ShowWQNodes       (DiagFlags_ & WaitQueueNodeBits)



//--- buffer for improving printf/etc performance on stdout ---
extern char far  outBuf[NUM_TASKS][StdOutBufSize];


/*-------------------------
  NOTES on pid's:
    The pid assigned to a process consists of 2 parts:
    - The least significant group of bits provides an index in to tasks[]
      (the # of bits controlled by the TaskNumMask macro)
    - The most significant group of bits provides a "usage" count - the
      # of times a given slot in tasks[] has been reused.
    This format avoids problems with a given pid being reused too soon
    (possibly confusing other processes trying to track a process via pid)
    while allowing the tasks[] entry for a process to be accessed very
    quickly without a search through the tasks[] (simply masking off the
    lower bits provides the index into tasks[])
----------------------------*/

#define  MyTaskNum   ( current - tasks )
#define  MyTaskIdx   ( (uint)MyTaskNum )

#define  TaskIdx(pid)  ((pid) & TaskIdxMask)
#define  Task_Idx(t)  ((uint)((t)-tasks))

PIX _activeTaskIdx_(PID pid);


extern uint schedCallsPrevSec, taskSwPrevSec;


#define  HostNameLen  32
extern char  hostName[HostNameLen];


extern bool  simulateMemCorruption;

extern uint  ctlLoopTicksPerSec, kernelTicksPerSec, saveTicksPerSec;

extern bool  checkWDT;  // do a dump if watchdog timer about to expire

extern char far osName[64], osDTM[32];

extern ulong  digitalIn, digitalOut, nextDigitalOut, prevDigitalOut;

extern DiagLogInfo  far prevLogInfo;

extern int  klocks;
extern bool  gotTime, pendingApplTick, startApplTicks,
             runBackup, runningPrimary;

extern uint  ttlTicks;  //tdebug
extern ulong ttlSecs;  //tdebug

extern bool  doReboot, doShutdown, useSoftReset, applicActive, restoredLog;

extern queueInfo  kernelQ;

extern FILE  taskIO, diagLog;

extern PIX  idlePix;

extern uchar  intTicks;
extern uint  milliTicks;
extern ulong  hdwResetTime, rebootTime;
extern uint applicTicks;

extern ulong  digitalIn, digitalOut;

extern ulong  deviceID;
extern uint  devType, hdwType, percentIdle;
extern int  timeZone;

extern char  moduleID[8];

extern unsigned long _farheapsize;


extern WaitQueue  sleepQueue;


#define  IN_CODE(addr)  \
    (((addr) >= codeStart) AND ((addr) <= codeEnd))

#define  IN_FARCODE(addr)  \
    (((addr) >= farCodeStart) AND ((addr) <= farCodeEnd))

#define  IN_ISR_TBL(addr)  \
    (((addr) >= intVectStart) AND ((addr) <= intVectEnd))


extern uint  codeStart, codeSize, codeEnd,
             constStart, constSize, constEnd,
             dataStart, dataSize, dataEnd,
             bssStart, bssSize, bssEnd;
extern ulong  farCodeStart, farCodeSize, farCodeEnd,
              farConstStart, farConstSize, farConstEnd,
              farBSSStart, farBSSSize, farBSSEnd;

extern uint  codeCRC, constCRC, farCodeCRC, farConstCRC;

extern  uint  intVectStart, intVectEnd;

extern uint  roStart, roEnd, wpStart, wpEnd;
extern ulong  roStartFar, roEndFar, wpStartFar, wpEndFar;

extern uint  wpSegMask, wpRootMask;

extern uchar  cpuID, cpuRev;
extern bool  hdwProtection;


void showMemProtection(void);

#define DisableWriteProtection()  { \
  outi(MMIDR, 0x80);  outi(WPCR, 0);  outi(MMIDR, 0x00);  }

#define EnableWriteProtection()  {  \
  outi(MMIDR, 0x80);  outi(WPCR, 1);  outi(MMIDR, 0x00); }


// these are in cstart
extern unsigned long _xmemSize, _xmemEnd, _xmemHeap;


void softReset(void);
void hardReset(void);


extern void postMortem(const char far *cause);


extern const far char asterisks[];

// deviceID's 0x00E0 - 0x00FF force a fixed IP addr of 192.168.250.<deviceID>
#define  FixedAddr() ((deviceID & 0xFFE0) == 0x00E0)


//--- values for devType (essentially the application type)---
#define  DT_BCM_RR        0x09   // BCM using Reed-Relay switches
#define  DT_BBServo       0x14   // Bang-Bang Servo
#define  DT_BCM_SS        0x19   // BCM using Solid-State switches
#define  DT_DCServo       0x1B   // DC Servo
#define  DT_RFPhase       0x20   // Analog RF Phase controller
#define  DT_FastRFOff     0x21   // Fast-acting RF Monitor using FPGA
#define  DT_RFServo       0x2B   // RF Servo controller
#define  DT_MWaveAmp      0x2C   // MicroWave Amplifier
#define  DT_ECROven       0x2D   // ECR Oven PS controller
#define  DT_RevSwitch     0x2E   // Reversing-Switch Power Supply controller
#define  DT_PhaseCtlr     0x2F   // Digital RF Phase controller
#define  DT_TCPS          0x30   // Trim-Coil PS Reversing Switch controller
#define  DT_RFMonitor     0x31   // RF Monitor (NIM module)
#define  DT_PS8Rack       0x50   // 8 supply rack-mount controller
#define  DT_ModularPS     0x51   // Modular Power Supply with LCD display
#define  DT_DVR           0x52   // D Voltage Regulator
#define  DT_SparkDetector 0x53   // Spark Detector NIM module
#define  DT_TCPS_2009     0x54   // DEBROS-based version of TCPS
#define  DT_LLRFTurnOn    0x55   // LLRF Turn On module
#define  DT_RFAC          0x56   // RF Amplifier Coordinator
#define  DT_LLRF2         0x57   // Low Level RF Ctlr 2
#define  DT_Starter       0xF8   // generic Starter application
#define  DT_Test          0xF9   // Various test commands for OS functions

//--- values for hdwType ---
#define  HT_UNKNOWN       0x00   // unknown/not set
#define  HT_BL2105        0x01   // ZWorld BL2105 SBC (uses RCM2100)
#define  HT_NSCL16        0x02   // NSCL 16-bit Analog SBC (uses RCM3200)
#define  HT_PHASE_CTLR    0x03   // NSCL Digital Phase Ctlr (uses RCM3200)
#define  HT_PS8_RACK      0x04   // NSCL 8-chan PS ctlr (uses RCM3200)
#define  HT_RFMON         0x05   // NSCL RF Monitor (uses RCM3200)
#define  HT_MODULAR_PS    0x06   // NSCL Modular PS (uses RCM3200)
#define  HT_RFServo       0x07   // NSCL RFServo NIM module (uses RCM3200)
#define  HT_DCServo       0x08   // NSCL DC Servo box (uses BL2105)
#define  HT_BBServo       0x09   // NSCL BB Servo box (uses BL2105)
#define  HT_SparkDetector 0x0A   // Spark Detector NIM module (uses RCM3200)
#define  HT_LLRFTurnOn    0x0B   // LLRF Turn ON module (uses RCM3200)
#define  HT_LLRF2         0x0C   // Low Level RF ctlr vers 2 (uses RCM3200)


//--- Rabbit 3000A registers not defined in Softools Rabbit.h ---
#define WPCR    0x0440   // Write Protection Control Register

#define STKCR   0x0444   // Stack Limit Control Register
#define STKLLR  0x0445   // Stack Low Limit Register
#define STKHLR  0x0446   // Stack High Limit Register

#define RAMSR   0x0448   // RAM Segment Register

#define WPLR    0x0460   // Write-Protect Low Register
#define WPHR    0x0461   // Write-Protect High Register

#define WPSAR   0x0480   // Write-Protect Segment A Register
#define WPSALR  0x0481   // Write-Protect Segment A Low Register
#define WPSAHR  0x0482   // Write-Protect Segment A High Register

#define WPSBR   0x0484   // Write-Protect Segment B Register
#define WPSBLR  0x0485   // Write-Protect Segment B Low Register
#define WPSBHR  0x0486   // Write-Protect Segment B High Register


//-----------------------------------------------------------------------------
//  some non-standard functions
//-----------------------------------------------------------------------------

void REBOOT(bool restartOS);

//--- returns 4 LSB bytes of 6-byte RTC clock counter ---
//--- (32768 ticks/sec: reading 4-bytes provides a 36.4 hour window) ---
ulong clockTicks(void);

int _stackFree_(PID pid);

uint  getPC(void);

PID createProc(ProcMain addr, char priority, int argc, char **argv,
               fd_t infd, fd_t outfd, fd_t errfd);

ulong sleepMS(ulong ms);

ulong msToTicks(ulong ms);

// A faster alternative to calling time(NULL) (?)
//    hdwResetTime is the time() value from the last CPU reset
//    _TimerSeconds is the # of seconds since hdwResetTime
#define  TIME()  ( hdwResetTime + _TimerSeconds )


char *statusDescr(PIX pix);

//void testPage0(char *when);  //tdebug


//--- functions that are actually in the modified xalloc.c ---
char far * bb_xalloc(unsigned long bb_size);  // actually in xalloc.c

extern ulong bbAllocatedCount;



extern char *Gets(char *str, char *history, uint idx, uint histSize);  //tdebug
#define  GETS(str)    Gets(str, NULL, 0, 0)
//#define  GETS(str)    fgets(str, sizeof(str), stdin)


void computeMemCRCs(bool firstTime, bool display);

void showIntVectors(void);


//-----------------------------------------------------------------------------
//--- standard Unix-compatible functions ---
//-----------------------------------------------------------------------------

int kill(PID pid, int sig);

PID waitpid(PID pid, int *status, int options);

void Exit(int stat);

int SetHostName(char *name, int nameLen);

int getpriority(int which, int who);
int setpriority(int which, int who, int priority);

int pipe(int fd[2]);


//--- macros for accessing elements of the process table ---
#define  errno  (current->lastErr)
#define  ERRNO  (current->lastErr)

#define  getpid()    ( current->pid )
#define  getppid()   ( current->ppid )



#define PRIO_PROCESS    0

//-----------------------------------------------------------------------------

#endif
