//----- _cmd_list.c ----- 07/30/12 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  The list of the standard command-line commands and ones that are indepent
//  of the application type.
//-----------------------------------------------------------------------------
//  UPDATE LOG (most recent first)
//
//
//  2012-02-14 - Mark Davis
//               - Changes to move commonCmds in to extended memory (much of it
//                 was still in root memory).
//
//  2012-02-09 - Mark Davis
//               - Changed shell command epics_cad to cas
//
//  2011-03-30 - Mark Davis
//               - Added nc command
//
//  2011-01-28 - Mark Davis
//               - Added route and ping commands
//
//  2010-05-25 - Mark Davis
//               - Added test_pv command
//
//  2010-05-?? - Mark Davis
//               - Added test_poll command
//
//  2009-12-09 - Mark Davis
//               - Added inttab command
//
//  2009-03-20 - Mark Davis
//               - placed list in to separate object file and each command
//                 in its own file in attempt to eliminate linkage problems.
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <sys/epics_cad.h>

#include <busybox.h>

#include <sys/backup.h>

#include <sys/hdwDefs.h>

#include <sys/srfs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

extern ulong _xmemSize;


#include <tsh.h>

#include <sys/modbus.h>

#include <sys/backup.h>

#include <applic.h>

#include <sys/farprintf.h>


extern void _task backupd(int argc, char **argv);
extern void _task configd(int argc, char **argv);
extern void _task inetd(int argc, char **argv);
extern void _task test_poll(int argc, char **argv);
extern void _task test_pv(int argc, char **argv);

//extern void _task timeClient(int argc, char **argv);


//-----------------------------------------------------------------------------

#define  CmdData(name, farFuncRedir, descr)  { \
  (ulong) (name),  \
  (uint)  (farFuncRedir),  \
  (ulong) (descr)  \
}

#pragma seg (CONST,FARCONST)

const far CmdInfo commonCmds[] = {
  { "adc",         _adc,           "Force/unforce value for ADC reading" },
  { "adccalib",    _adccalib,      "Ctlr-specific Analog input calibrations" },
  { "anares",      _anares,        "Show current res of analog read/set signals" },
  { "args",        _args,          "Test parsing of cmd line arguments" },
  { "arp",         _arp,           "Show arp table" },
  { "backupd",     backupd,        "Automatic backup/restore daemon" },
  { "bootimg",     _bootimg,       "Show/Set Primary/Backup boot image" },
  { "btof",        _btof,          "Convert old blocks to files" },
  { "cas",         cas,            "manage EPICS CA protocol server daemon" },
  { "cat",         _cat,           "Copy file to stdout" },
  { "cls",         NULL,           "Clear the screen" },
  { "configd",     configd,        "Automatic Configuration daemon" },
  { "cp",          _cp,            "Copy a file" },
  { "dac",         _dac,           "Force/unforce value for DAC output" },
  { "daccalib",    _daccalib,      "Ctlr-specific Analog output calibrations" },
  { "date",        _date,          "show date/time" },
  { "del",         _rm,            "remove (delete) a file" },
  { "dig",         _dig,           "Digital I/O cmds" },
  { "df",          _df,            "show free space on file system" },
  { "dumpport",    _dumpPort,      "show values from serial port struct" },
  { "eng",         _eng,           "test Engineering() floating formater" },
  { "diag",        _diag,          "Show/Set diagnostic output flags" },
  { "dir",         _ls,            "List files" },
  { "exit",        NULL,           "Exit shell" },
  { "fifoClient",  _fifoClient,    "FIFO test client" },
  { "fifoServ",    _fifoServ,      "FIFO test Server" },
  { "float",       _float,         "Test of alternate floating point funcs" },
  { "free",        _free,          "How much xmem is avail" },
  { "fsck",        _fsck,          "check integrity of the file system" },
  { "getty",       getty,          "Start login shell on serial device" },
  { "grep",        _grep,          "Show lines containing specified string" },
  { "hang",        _hang,          "Test WatchDog timer reboot" },
  { "help",        _help,          "List commands" },
  { "hex",         _od,            "show data in hex format" },
  { "hostname",    __hostname,     "set/show hostname" },
  { "ifconfig",    _ifconfig,      "Show/set network interface config" },
  { "inetd",       inetd,          "Internet 'super-server'" },
  { "inttab",      _inttab,        "Interrupt Table information" },
  { "ios",         _ios,           "Digital/Analog I/O status" },
  { "kill",        _kill,          "Send a Signal to a process" },
  { "label",       _label,         "Set/show input/output signal labels" },
  { "ls",          _ls,            "List files" },
  { "lsof",        _lsof,          "List open file descriptors" },
//  { "ltest",       _ltest,         "List test" },
  { "more",        _more,          "show a file or stdin a page at a time" },
  { "mv",          _mv,            "move (rename) a file" },
  { "nc",          _nc,            "connect stdio to remote port" },
  { "netstat",     _netstat,       "Show network connections" },
  { "od",          _od,            "Octal Dump (copy/translate data)" },
  { "ping",        _ping_,         "Send ICMP ECHO_REQUEST to network hosts" },
  { "ps",          _ps,            "Process Status" },
  { "rdflash",     _rdflash,       "Read data from flash" },
  { "rdmem",       _rdmem,         "Read data from memory" },
//  { "rduserblk",   _rduserblk,     "Read data from User Block" },
  { "readcalib",    _readcalib,    "Device-specific Analog input calibrations" },
  { "reboot",      _reboot,        "Shutdown and restart OS" },
  { "reg",         _reg,           "List or modify Modbus Registers" },
  { "reload",      paramFileCmd,   "Reload 'persistent' values from file(s)" },
  { "renice",      _renice,        "Change priority of running process" },
  { "rm",          _rm,            "Remove (delete) a file" },
  { "route",       _route,         "Show / manipulate the IP routing table" },
  { "save",        paramFileCmd,   "save 'persistent' values to file(s)" },
  { "setcalib",    _setcalib,      "Device-specific Analog output calibrations" },
  { "scope",       _scope,         "Capture/display waveform data" },
  { "sh",          _sh,            "Run commands from a text file" },
  { "shutdown",    _shutdown,      "Shutdown and run Download Manager" },
  { "srfs",        _srfs,          "Set/Show SRFS service parameters" },
  { "stat",        _stat,          "Show file status" },
  { "sysinfo",     _sysinfo,       "Hardware/System information" },
  { "syslog",      _syslog,        "Show new diagnostic/system log" },
  { "testOut",     _testOut,       "Test continuous output" },
  { "test_poll",   test_poll,      "Test poll() function" },
  { "test_pv",     test_pv,        "Test PV interface to registers" },
  { "testgets",    _testgets,      "Test gets()" },
  { "tip",         _tip,           "Terminal Interface Program" },
  { "top",         _top,           "Continuous process status" },
  { "unlink",      _rm,            "remove (delete) a file" },
  { "uname",       _uname,         "Description of Operating System" },
  { "update",      _update,        "Update Boot Image file" },
  { "watch",       _watch,         "Monitor the value of one or more regs" }
};

#pragma seg (CONST)

int const NumCommonCmds = ( sizeof(commonCmds) / sizeof(CmdInfo) );


//-----------------------------------------------------------------------------
