//----- hang.c ----- 08/09/13 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//
//  2013-08-09 - Mark Davis
//               - Added "mem" option to simulate corruption of memory,

//  2011-03-10 - Mark Davis
//               - The "write" option now uses a critical section rather than
//                 the scheduler lock.
//
//  2010-06-07 - Mark Davis
//               - Added 'stack' option to cause a stack overflow
//
//  2010-06-02 - Mark Davis
//               - Added 'write' option to test memory fault/write protection
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <tsh.h>

#include <applic.h>

#include <sys/farprintf.h>

//-----------------------------------------------------------------------------
//  Get close enough to an overflow to invoke the hdw protection but not
//  (if possible) actually overflow the stack.
//-----------------------------------------------------------------------------
void stackOverflow(void)
{
  if (_stackFree_(current->pid) < 14)  return;

  stackOverflow();
}

//-----------------------------------------------------------------------------
//  Purposely disable scheduler to test auto-reboot by WatchDog timer
//-----------------------------------------------------------------------------
extern void cstart(void);

void _task _hang(int argc, char **argv)
{
  ulong  addr;
  uint  val;


  if (showHelp(2, argc, argv))  {
    printf("\nusage: hang mem|lock|jump0|cstart|stack|write <addr>\r\n");  Exit(-1); }

  if (StriEq(argv[1], "mem"))  {
    printf("Hanging by pretending memory has been corrupted...\r\n");
    sleepMS(1000);  simulateMemCorruption = true;  Exit(0);
  }

  if (StriEq(argv[1], "lock"))  {
    printf("Hanging by doing infinite loop with sched disabled...\r\n");
    sleepMS(1000);  KLOCK();  for(;;)  { }
  }

  if (StriEq(argv[1], "jump0"))  {
    printf("Jumping to address 0...\r\n");
    sleepMS(1000);  _asm( "jp 0" );
  }

  if (StriEq(argv[1], "cstart"))  {
    printf("Jumping to address 'cstart'...\r\n");
    sleepMS(1000);   cstart();
  }

  if (StriEq(argv[1], "stack"))  stackOverflow();

  if (StriEq(argv[1], "write"))  {
    if (argc < 3)  { printf("addr not specified\r\n");  Exit(-1);  }
    addr = ATOL(argv[2]);
    printf("Writing to location 0x%05lX...\r\n", addr);
    _EnterCritical_();
    val = *((uint far *)addr);
    *((uint far *)addr) = ~val;  *((uint far *)addr) = val;
    _ExitCritical_();
    Exit(0);
  }


  printf("Unknown option: %.32s\r\n", argv[1]);  Exit(-1);
}

//-----------------------------------------------------------------------------
