//----- lsof.c ----- 03/15/11 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//
//  2011-03-15 - Mark Davis
//               - changed -counts option to -c and added -r option to
//                 keep repeating the standard-style lsof info
//
//  2009-06-12 - Mark Davis
//               - added FLAGS column and removed dummy USER column
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <tsh.h>

#include <applic.h>

#include <sys/farprintf.h>

//-----------------------------------------------------------------------------
//  list open files
//-----------------------------------------------------------------------------
void _task _lsof(int argc, char **argv)
{
#define MaxDescArgs  8
  fd_t   fd;
  uchar  i, j, pix;
  FDEV  far *file;
  bool  showCounts, repeat, hasOpenFDs;
  void  *ioDev;
  task_t  *task;
  char  *argv_[MaxDescArgs];
  int    argc_;
  char   buf[200];

  static const char far StdCols[] = {
 "COMMAND     PID  PFD GFD   FLAGS     TYPE  DEVICE       NODE NAME" };
//xxxxxxxxx xxxxx  xxx xxx  0x####  xxxxxxx  xxxxxx  xxxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxxxxx


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  lsof [-c|-r]\r\n");  Exit(-1); }

  showCounts = repeat = false;
  if (argc > 1)  for (i=1; i<argc; ++i)  {
    if (strcmp(argv[i], "-c") == 0)  showCounts = true;
    if (strcmp(argv[i], "-r") == 0)  repeat = true;
  }


  if (showCounts OR repeat)  printf(clsCmd);

  if (showCounts)  for (;;)  {
    stdHeader();  printf("%s\n", cllCmd);
    printf("     NODE        SENT        RCVD  NAME%s\r\n", cllCmd);
    for (i=0; i<MaxOpenFiles; ++i)  {
      if ((file = fileTbl + i)->openCount == 0)  continue;
      ioDev = file->ioDev;
      file->ioFuncs->descr(ioDev, buf, sizeof(buf)-1);
      parseArgs(buf, MaxDescArgs, argv_, &argc_);
      printf("%9.9s  %10lu  %10lu ",
             argv_[2], file->writeCount, file->readCount);
      for (j=3; j<argc_; ++j)  printf(" %s", argv_[j]);
      printf("%s\r\n", cllCmd);
    }
    printf(clpCmd);  fflush(stdout);
    CheckETX();  sleepMS(1000);
  }


/*
===== Sample from a Linux machine =====

COMMAND     PID   USER   FD      TYPE     DEVICE      NODE NAME
init          1   root  cwd       DIR        3,1         2 /
init          1   root  rtd       DIR        3,1         2 /
init          1   root  txt       REG        3,1        49 /sbin/init
init          1   root  mem       REG        3,1     14777 /lib/tls/libc-2.3.2.so
portmap     765 daemon    2u      CHR        1,3     43976 /dev/null
portmap     765 daemon    3u     IPv4       2924       UDP *:sunrpc
portmap     765 daemon    4u     IPv4       2925       TCP *:sunrpc (LISTEN)
winbindd    947   root   10ur     REG        3,5     48015 /var/run/samba/messages.tdb
rpciod      961   root  txt   unknown                      /proc/961/exe
QtChan    16796  epics    7u     unix 0xf5ee6ac0   3922134 socket
xxxxxxxxx xxxxx xxxxxx xxxx-- xxxxxxx xxxxxxxxxx xxxxxxxxx xxxxxxxxxxxxxxxxxxxxxxxxxxx
AAAAAAAAA ##### daemon   ##      IPv4 0x####[##]       TCP ###.###.###.###:#####->###.###.###.###:#####


===== Sample from code below =====

COMMAND     PID  PFD GFD   FLAGS     TYPE  DEVICE       NODE NAME
[kernel]      0    3   0  0x0002      spi  0x5651     ser[1] /dev/spi1 mode:1 rate:1000000

applic        2    3   1  0x0021     fifo  0x5626    fifo[7] fifoEvents
applic        2    4   2  0x0002      spi  0x5669     ser[2] /dev/spi2 mode:0 rate:1000000

inetd         3    3   6  0x0022     IPv4  0x62EF     TCP[2] *:2004 -> *:* Listen
inetd         3    4   7  0x0002   serial  0x5639     ser[0] /dev/tty0 115200 baud
inetd         3    5   8  0x0022     IPv4  0x6319     TCP[3] *:2003 -> *:* Listen
inetd         3    6   9  0x0022     IPv4  0x629B     TCP[0] 192.168.125.135:502 -> *:* Listen
inetd         3    7  10  0x0022     IPv4  0x62C5     TCP[1] 192.168.125.135:23 -> *:* Listen

configd       4    3   5  0x0022     IPv4  0x6493     UDP[0] *:2001 -> 255.255.255.255:* UDP Socket

backupd       5    3   3  0x0002     fifo  0x5626    fifo[7] fifoEvents
backupd       5    4  11  0x0016     srfs  0x6DF2    srfs[0] PhCtlr_x00_06-06.log

tsh           6    0  12  0x0002     IPv4  0x636D     TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Listen
tsh           6    1  12  0x0002     IPv4  0x636D     TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Listen
tsh           6    2  12  0x0002     IPv4  0x636D     TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Listen

epics_cad    42    3  14  0x0022     IPv4  0x64BD     UDP[1] *:5064 -> 255.255.255.255:5065 UDP Socket
epics_cad    42    4  15  0x0022     IPv4  0x6469    TCP[11] 192.168.125.135:5064 -> *:* Listen
epics_cad    42    5  16  0x0002     IPv4  0x6415     TCP[9] 192.168.125.135:5064 -> 192.168.4.5:47776 Listen
epics_cad    42    6  17  0x0002     IPv4  0x643F    TCP[10] 192.168.125.135:5064 -> 192.168.4.6:47223 Listen

top          75    0  12  0x0002     IPv4  0x636D     TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Listen
top          75    1  12  0x0002     IPv4  0x636D     TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Listen
top          75    2  12  0x0002     IPv4  0x636D     TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Listen

modbusTCP    76    0  18  0x0002     IPv4  0x6343     TCP[4] 192.168.125.135:502 -> 192.168.4.28:34434 Listen
modbusTCP    76    1  18  0x0002     IPv4  0x6343     TCP[4] 192.168.125.135:502 -> 192.168.4.28:34434 Listen
modbusTCP    76    2  18  0x0002     IPv4  0x6343     TCP[4] 192.168.125.135:502 -> 192.168.4.28:34434 Listen
modbusTCP    76    3  19  0x0002     fifo  0x5626    fifo[7] fifoEvents

tsh          14    0  13  0x0002     IPv4  0x63C1     TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Listen
tsh          14    1  13  0x0002     IPv4  0x63C1     TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Listen
tsh          14    2  13  0x0002     IPv4  0x63C1     TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Listen

lsof         82    0  13  0x0002     IPv4  0x63C1     TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Listen
lsof         82    1  13  0x0002     IPv4  0x63C1     TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Listen
lsof         82    2  13  0x0002     IPv4  0x63C1     TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Listen
*/

  for (;;)  {
    if (repeat)  stdHeader();
    printf("%s\r\n", cllCmd);  printf(StdCols);  printf("%s\r\n", cllCmd);

    for (pix=0; pix<NUM_TASKS; ++pix)  if ((task = tasks + pix)->inUse)  {
      hasOpenFDs = false;
      for (fd=0; fd<FD_SETSIZE; ++fd)  {
        if (! (file = FD[pix][fd]))  continue;
        ioDev = file->ioDev;
        file->ioFuncs->descr(ioDev, buf, sizeof(buf)-1);
        for (i=0; i<MaxDescArgs; ++i)  argv_[i] = " ";
        parseArgs(buf, MaxDescArgs, argv_, &argc_);
        printf("%-9.9s %5u  %3d %3u  0x%04X %8.7s  %6.6s  %9.9s",
               task->argv[0], task->pid, fd, File_ID(file), file->flags,
               argv_[0], argv_[1], argv_[2] );
        for (i=3; i<MaxDescArgs; ++i)  printf(" %s", argv_[i]);
        printf("%s\r\n", cllCmd);  hasOpenFDs = true;
      }
      if (hasOpenFDs)  printf("%s\r\n", cllCmd);
    }

/*
===== from fileSysCheck(): Status by file system =====
checking fsf[0] (fifo)
  [ 7]:  3 refs:  fifo 0x5626 fifo[7] fifoEvents

checking fsf[1] (socket)
  [ 0]:  1 refs:  IPv4 0x629B TCP[0] 192.168.125.135:502 -> *:* Listen
  [ 1]:  1 refs:  IPv4 0x62C5 TCP[1] 192.168.125.135:23 -> *:* Listen
  [ 2]:  1 refs:  IPv4 0x62EF TCP[2] *:2004 -> *:* Listen
  [ 3]:  1 refs:  IPv4 0x6319 TCP[3] *:2003 -> *:* Listen
  [ 4]:  1 refs:  IPv4 0x6343 TCP[4] 192.168.125.135:502 -> 192.168.4.28:34434 Established
  [ 5]:  1 refs:  IPv4 0x636D TCP[5] 192.168.125.135:23 -> 192.168.100.54:58067 Established
  [ 6]:  1 refs:  IPv4 0x6397 TCP[6] 192.168.125.135:1026 -> 192.168.4.34:2002 Established
  [ 7]:  1 refs:  IPv4 0x63C1 TCP[7] 192.168.125.135:23 -> 192.168.100.54:58169 Established
  [ 9]:  1 refs:  IPv4 0x6415 TCP[9] 192.168.125.135:5064 -> 192.168.4.5:47776 Established
  [10]:  1 refs:  IPv4 0x643F TCP[10] 192.168.125.135:5064 -> 192.168.4.6:47223 Established
  [11]:  1 refs:  IPv4 0x6469 TCP[11] 192.168.125.135:5064 -> *:* Listen
  [12]:  1 refs:  IPv4 0x6493 UDP[0] *:2001 -> 255.255.255.255:* UDP Socket
  [13]:  1 refs:  IPv4 0x64BD UDP[1] *:5064 -> 255.255.255.255:5065 UDP Socket

checking fsf[2] (srfs)
  [ 0]:  1 refs:  srfs 0x6DF2 srfs[0] PhCtlr_x00_06-06.log

checking fsf[3] (serial)
  [ 0]:  1 refs:  serial 0x5639 ser[0] /dev/tty0 115200 baud
  [ 1]:  1 refs:  serial 0x5651 ser[1] /dev/tty1 0 baud
  [ 2]:  1 refs:  serial 0x5669 ser[2] /dev/tty2 0 baud
*/
    fileSysCheck(false, buf, sizeof(buf));  // check for orphaned or corrupt fds
    printf(clpCmd);  fflush(stdout);

    if (!repeat)  break;
    CheckETX();  sleepMS(2000);
  }



  Exit(0);
}

//-----------------------------------------------------------------------------
