//----- test_poll.c ----- 01/20/12 -----(05/12/10)-----

//-----------------------------------------------------------------------------
//  process to test poll function
//-----------------------------------------------------------------------------
// Update Log (most recent first)
//
//  Entries marked with (*) indicate ones that improve the stability and/or
//  reliability of the code.
//
//
//  2012-01-20 - Mark Davis
//               - Changed path for poll.h
//
//  2010-05-21 - Mark Davis
//               - first version completed and tested
//
//-----------------------------------------------------------------------------
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <poll.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>
#include <linux/sched.h>
#include <errno.h>

#include <string.h>

#include <stcpip.h>
#include <inetd.h>


#include <sys/farprintf.h>

#include <tsh.h>

#include <busybox.h>



#define MaxClients  3


ServiceInfo srv = { "pollTest", 0, ProtoTCP, NULL,  10, -1, NULL, 0, 0 };

//-----------------------------------------------------------------------------
//  process to listen on specified ports and create processes to service
//  incoming connections on those ports
//-----------------------------------------------------------------------------
void _task test_poll(int argc, char **argv)
{
  fd_t  devfd, newDevfd;
  uint  u, numClients;
  struct pollfd  *client, clients[MaxClients];
  int  errNum, addrLen, stat;
  struct sockaddr_in  cliAddr;
  FDEV  far *file;
  void  *ioDev;
  char  temp[128];


//  redirectStdio();  // so printf, perror, etc don't end up in data stream

  if (showHelp(2, argc, argv))  {
    printf("\nusage: test_poll port\r\n\n");  Exit(-1); }

  srv.port = atoi(argv[1]);  srv.devfd = -1;

  for (u=0; u<MaxClients; ++u)  {
    clients[u].fd = -1;  clients[u].events = POLLIN; }

  numClients = 0;

  for (;;)  {
    CheckETX();

  //--- make sure we have a listening socket ---
    if ((devfd = srv.devfd) < 0)  {
      getListener(&srv);  sleepMS(100);  continue; }

  //--- check for a new connection ---
    addrLen = sizeof(cliAddr);
    newDevfd = accept(devfd, (struct sockaddr *)&cliAddr, &addrLen);

    if (newDevfd >= 0)  {
      file = FD[userPix][newDevfd];  ioDev = file->ioDev;
      file->ioFuncs->descr(ioDev, temp, sizeof(temp)-1);
      printf("opened %.32s connection on fd: %d (GFD: %d)\r\n%.128s\r\n\n",
             srv.srvName, newDevfd, GFD(newDevfd), temp );

      if (numClients > MaxClients)  {
        printf("Max # clients exceeded\r\n");  close(newDevfd);  }
      else  {
        printf("new client on fd %d\r\n", newDevfd);
        clients[numClients].fd = newDevfd;
        ++numClients;
      }
    }
    else  if (errno != EAGAIN)  {
      errNum = errno;
      printf("accept() for %.32s: fd: %d (GFD %d)\r\nError %d: ",
             srv.srvName, devfd, GFD(devfd), errNum);
      perror(NULL);
      close(devfd);  srv.devfd = -1;
    }

    if (!numClients)  continue;


  //--- block for a bit waiting for input from one or more of the clients ---
    stat = poll(clients, numClients, 100);

    if (!stat)  continue;  // no errors, nothing new to read

    if (stat < 0)  { perror("test_poll/poll()");  errno = 0; }

  //--- check to see what (if any) clients sent us something or if any ---
  //--- errors occured ---
    for (u=0; u<numClients; ++u)  {
      client = clients + u;

      do  {
        if (client->revents & (POLLERR | POLLNVAL | POLLHUP))  {
          if (client->revents & POLLERR)
            printf("[%u] error on connection\r\n", client->fd);
          if (client->revents & POLLNVAL)
            printf("[%u] invalid fd\r\n", client->fd);
          if (client->revents & POLLHUP)
            printf("[%u] connection broken\r\n", client->fd);

          printf("[%u] closing connection\r\n", client->fd);

          close(client->fd);  client->fd = -1;

          if (u >= --numClients)  break;

          *client = clients[numClients];  continue;
        }

        if (client->revents & POLLIN)  {
          stat = read_(client->fd, temp, sizeof(temp)-1, 10);
          if (stat > 0)  {
            temp[stat] = 0;
            printf("[%u]: [%s]\r\n", client->fd, temp);
          }
        }

      } while(0);
    }

  }

}

//-----------------------------------------------------------------------------
